"""
JSON
----

Functions for reading from and writing to disk records in JSON format,
as one record per file or one record per *line* in a file.
"""
import datetime
import functools
import json
import pathlib
from typing import Any, Iterable, Optional, Tuple, Union

from . import utils as io_utils


def read_json(
    filepath: Union[str, pathlib.Path],
    *,
    mode: str = "rt",
    encoding: Optional[str] = None,
    lines: bool = False,
) -> Iterable:
    """
    Read the contents of a JSON file at ``filepath``, either all at once
    or streaming item-by-item.

    Args:
        filepath: Path to file on disk from which data will be read.
        mode: Mode with which ``filepath`` is opened.
        encoding: Name of the encoding used to decode or encode the data
            in ``filepath``. Only applicable in text mode.
        lines: If False, all data is read in at once; otherwise, data is read in
            one line at a time.

    Yields:
        Next JSON item; could be a dict, list, int, float, str,
        depending on the data and the value of ``lines``.
    """
    io_utils._validate_read_mode(mode)
    with io_utils.open_sesame(filepath, mode=mode, encoding=encoding) as f:
        if lines is False:
            yield json.load(f)
        else:
            for line in f:
                yield json.loads(line)


def read_json_mash(
    filepath: Union[str, pathlib.Path],
    *,
    mode: str = "rt",
    encoding: Optional[str] = None,
    buffer_size: int = 2048,
) -> Iterable:
    """
    Read the contents of a JSON file at ``filepath`` one item at a time,
    where all of the items have been mashed together, end-to-end, on a single line.

    Args:
        filepath: Path to file on disk to which data will be written.
        mode: Mode with which ``filepath`` is opened.
        encoding: Name of the encoding used to decode or encode the data
            in ``filepath``. Only applicable in text mode.
        buffer_size: Number of bytes to read in as a chunk.

    Yields:
        Next valid JSON object, converted to native Python equivalent.

    Note:
        Storing JSON data in this format is Not Good. Reading it is doable, so
        this function is included for users' convenience, but note that
        there is no analogous ``write_json_mash()`` function. Don't do it.
    """
    io_utils._validate_read_mode(mode)
    json_decoder = json.JSONDecoder()
    with io_utils.open_sesame(filepath, mode=mode, encoding=encoding) as f:
        buffer_ = ""
        for chunk in iter(functools.partial(f.read, buffer_size), ""):
            buffer_ += chunk
            while buffer_:
                try:
                    result, index = json_decoder.raw_decode(buffer_)
                    yield result
                    buffer_ = buffer_[index:]
                # not enough data to decode => read another chunk
                except ValueError:
                    break


def write_json(
    data: Any,
    filepath: Union[str, pathlib.Path],
    *,
    mode: str = "wt",
    encoding: Optional[str] = None,
    make_dirs: bool = False,
    lines: bool = False,
    ensure_ascii: bool = False,
    separators: Tuple[str, str] = (",", ":"),
    sort_keys: bool = False,
    indent: Optional[Union[int, str]] = None,
) -> None:
    """
    Write JSON ``data`` to disk at ``filepath``, either all at once
    or streaming item-by-item.

    Args:
        data: JSON data to write to disk, including any Python objects
            encodable by default in :mod:`json`, as well as dates and datetimes.
            For example::

                [
                    {"title": "Harrison Bergeron", "text": "The year was 2081, and everybody was finally equal."},
                    {"title": "2BR02B", "text": "Everything was perfectly swell."},
                    {"title": "Slaughterhouse-Five", "text": "All this happened, more or less."},
                ]

            If ``lines`` is False, all of ``data`` is written as a single object;
            if True, each item is written to a separate line in ``filepath``.

        filepath: Path to file on disk to which data will be written.
        mode: Mode with which ``filepath`` is opened.
        encoding: Name of the encoding used to decode or encode the data
            in ``filepath``. Only applicable in text mode.
        make_dirs: If True, automatically create (sub)directories if
            not already present in order to write ``filepath``.
        lines: If False, all data is written at once;
            otherwise, data is written to disk one item at a time.
        ensure_ascii: If True, all non-ASCII characters are escaped;
            otherwise, non-ASCII characters are output as-is.
        separators: An (item_separator, key_separator) pair
            specifying how items and keys are separated in output.
        sort_keys: If True, each output dictionary is sorted by key;
            otherwise, dictionary ordering is taken as-is.
        indent: If a non-negative integer or string, items are pretty-printed
            with the specified indent level; if 0, negative, or "", items are separated
            by newlines; if None, the most compact representation is used
            when storing ``data``.

    See Also:
        https://docs.python.org/3/library/json.html#json.dump
    """
    io_utils._validate_write_mode(mode)
    with io_utils.open_sesame(filepath, mode=mode, encoding=encoding, make_dirs=make_dirs) as f:
        if lines is False:
            f.write(
                json.dumps(
                    data,
                    indent=indent,
                    ensure_ascii=ensure_ascii,
                    separators=separators,
                    sort_keys=sort_keys,
                    cls=ExtendedJSONEncoder,
                )
            )
        else:
            newline = "\n" if "t" in mode else b"\n"
            for item in data:
                f.write(
                    json.dumps(
                        item,
                        indent=indent,
                        ensure_ascii=ensure_ascii,
                        separators=separators,
                        sort_keys=sort_keys,
                        cls=ExtendedJSONEncoder,
                    )
                    + newline
                )


class ExtendedJSONEncoder(json.JSONEncoder):
    """
    Sub-class of :class:`json.JSONEncoder`, used to write JSON data to disk in
    :func:`write_json()` while handling a broader range of Python objects.

    - :class:`datetime.datetime` => ISO-formatted string
    - :class:`datetime.date` => ISO-formatted string
    """

    def default(self, obj):
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        else:
            return super().default(obj)



import itertools
import random
from typing import cast, List, Optional, Set, Union

from cytoolz import itertoolz

from .. import utils
from . import utils as aug_utils


def substitute_word_synonyms(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
    pos: Optional[Union[str, Set[str]]] = None,
) -> List[aug_utils.AugTok]:
    """
    Randomly substitute words for which synonyms are available
    with a randomly selected synonym,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through synonym substitution.
        num: If int, maximum number of words with available synonyms
            to substitute with a randomly selected synonym; if float, probability
            that a given word with synonyms will be substituted.
        pos: Part of speech tag(s) of words to be considered for augmentation.
            If None, all words with synonyms are considered.

    Returns:
        New, augmented sequence of tokens.

    Note:
        This transform requires :class:`textacy.resources.ConceptNet` to be downloaded
        to work properly, since this is the data source for word synonyms to be substituted.
    """
    _validate_aug_toks(aug_toks)
    pos = cast(Set[str], utils.to_collection(pos, str, set))
    cand_idxs = [
        idx for idx, aug_tok in enumerate(aug_toks)
        if aug_tok.syns and (pos is None or aug_tok.pos in pos)
    ]
    rand_idxs = set(_select_random_candidates(cand_idxs, num))
    if not rand_idxs:
        return aug_toks[:]

    new_aug_toks = []
    for idx, aug_tok in enumerate(aug_toks):
        if idx in rand_idxs:
            new_aug_toks.append(
                aug_utils.AugTok(
                    text=random.choice(aug_tok.syns),
                    ws=aug_tok.ws,
                    pos=aug_tok.pos,
                    is_word=aug_tok.is_word,
                    syns=aug_tok.syns,  # TODO: re-fetch syns? use []?
                )
            )
        else:
            new_aug_toks.append(aug_tok)
    return new_aug_toks


def insert_word_synonyms(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
    pos: Optional[Union[str, Set[str]]] = None,
) -> List[aug_utils.AugTok]:
    """
    Randomly insert random synonyms of tokens for which synonyms are available,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through synonym insertion.
        num: If int, maximum number of words with available synonyms
            from which a random synonym is selected and randomly inserted; if float,
            probability that a given word with synonyms will provide a synonym
            to be inserted.
        pos: Part of speech tag(s) of words to be considered for augmentation.
            If None, all words with synonyms are considered.

    Returns:
        New, augmented sequence of tokens.

    Note:
        This transform requires :class:`textacy.resources.ConceptNet` to be downloaded
        to work properly, since this is the data source for word synonyms to be inserted.
    """
    _validate_aug_toks(aug_toks)
    pos = cast(Set[str], utils.to_collection(pos, str, set))
    # bail out on very short sentences to avoid clobbering meaning
    if len(aug_toks) < 3:
        return aug_toks[:]

    cand_aug_toks = [
        aug_tok for aug_tok in aug_toks
        if aug_tok.syns and (pos is None or aug_tok.pos in pos)
    ]
    rand_aug_toks = _select_random_candidates(cand_aug_toks, num)
    rand_idxs = random.sample(range(len(aug_toks)), len(rand_aug_toks))
    if not rand_idxs:
        return aug_toks[:]

    rand_aug_toks = iter(rand_aug_toks)
    new_aug_toks: List[aug_utils.AugTok] = []
    for idx, (prev_tok, curr_tok) in enumerate(itertoolz.sliding_window(2, [None] + aug_toks)):
        if idx in rand_idxs:
            rand_aug_tok = next(rand_aug_toks)
            if prev_tok:
                # use previous token's whitespace for inserted synonym
                new_tok_ws = prev_tok.ws
                if prev_tok.is_word and not prev_tok.ws:
                    # previous token should have whitespace, if a word
                    new_aug_toks[-1] = aug_utils.AugTok(
                        text=prev_tok.text,
                        ws=" ",
                        pos=prev_tok.pos,
                        is_word=True,
                        syns=prev_tok.syns,
                    )
            else:
                new_tok_ws = " "
            new_aug_toks.append(
                aug_utils.AugTok(
                    text=random.choice(rand_aug_tok.syns),
                    ws=new_tok_ws,
                    pos=rand_aug_tok.pos,
                    is_word=rand_aug_tok.is_word,
                    syns=rand_aug_tok.syns, # TODO: re-fetch syns? use []?
                )
            )
        new_aug_toks.append(curr_tok)
    return new_aug_toks


def swap_words(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
    pos: Optional[Union[str, Set[str]]] = None,
) -> List[aug_utils.AugTok]:
    """
    Randomly swap the positions of two *adjacent* words,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through position swapping.
        num: If int, maximum number of adjacent word pairs to swap;
            if float, probability that a given word pair will be swapped.
        pos: Part of speech tag(s) of words to be considered for augmentation.
            If None, all words are considered.

    Returns:
        New, augmented sequence of tokens.
    """
    _validate_aug_toks(aug_toks)
    pos = cast(Set[str], utils.to_collection(pos, str, set))
    # if we don't require _adjacent_ words, this does the trick
    # if not pos:
    #     pos = set(aug_tok.pos for aug_tok in aug_toks if aug_tok.is_word)
    # cand_idx_pairs = list(
    #     itertools.chain.from_iterable(
    #         itertools.combinations(
    #             (idx for idx, aug_tok in enumerate(aug_toks) if aug_tok.pos == pos_),
    #             2,
    #         )
    #         for pos_ in pos
    #     )
    # )
    cand_idxs = (
        idx for idx, aug_tok in enumerate(aug_toks)
        if aug_tok.is_word and (pos is None or aug_tok.pos in pos)
    )
    cand_idx_pairs = [
        (idx1, idx2) for idx1, idx2 in itertoolz.sliding_window(2, cand_idxs)
        if idx2 - idx1 == 1
    ]
    rand_idx_pairs = _select_random_candidates(cand_idx_pairs, num)
    if not rand_idx_pairs:
        return aug_toks[:]

    new_aug_toks = aug_toks[:]
    for idx1, idx2 in rand_idx_pairs:
        tok1 = new_aug_toks[idx1]
        tok2 = new_aug_toks[idx2]
        new_aug_toks[idx1] = aug_utils.AugTok(
            text=tok2.text,
            ws=tok1.ws,
            pos=tok2.pos,
            is_word=tok2.is_word,
            syns=tok2.syns,
        )
        new_aug_toks[idx2] = aug_utils.AugTok(
            text=tok1.text,
            ws=tok2.ws,
            pos=tok1.pos,
            is_word=tok1.is_word,
            syns=tok1.syns,
        )
    return new_aug_toks


def delete_words(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
    pos: Optional[Union[str, Set[str]]] = None,
) -> List[aug_utils.AugTok]:
    """
    Randomly delete words,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through word deletion.
        num: If int, maximum number of words to delete;
            if float, probability that a given word will be deleted.
        pos: Part of speech tag(s) of words to be considered for augmentation.
            If None, all words are considered.

    Returns:
        New, augmented sequence of tokens.
    """
    _validate_aug_toks(aug_toks)
    pos = cast(Set[str], utils.to_collection(pos, str, set))
    # bail out on very short sentences to avoid clobbering meaning
    if len(aug_toks) < 3:
        return aug_toks[:]

    cand_idxs = [
        idx for idx, aug_tok in enumerate(aug_toks)
        if aug_tok.is_word and (pos is None or aug_tok.pos in pos)
    ]
    rand_idxs = set(_select_random_candidates(cand_idxs, num))
    if not rand_idxs:
        return aug_toks[:]

    new_aug_toks: List[aug_utils.AugTok] = []
    padded_triplets = itertoolz.sliding_window(3, [None] + aug_toks + [None])
    for idx, (prev_tok, curr_tok, next_tok) in enumerate(padded_triplets):
        if idx in rand_idxs:
            # special case: word then [deleted word] then punctuation
            # give deleted word's whitespace to previous word
            if prev_tok and next_tok and prev_tok.is_word and not next_tok.is_word:
                new_aug_toks[-1] = aug_utils.AugTok(
                    text=prev_tok.text,
                    ws=curr_tok.ws,
                    pos=prev_tok.pos,
                    is_word=prev_tok.is_word,
                    syns=prev_tok.syns,
                )
        else:
            new_aug_toks.append(curr_tok)
    return new_aug_toks


def substitute_chars(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
    lang: Optional[str] = None,
) -> List[aug_utils.AugTok]:
    """
    Randomly substitute a single character in randomly-selected words with another,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through character substitution.
        num: If int, maximum number of words to modify with a random character substitution;
            if float, probability that a given word will be modified.
        lang: Standard, two-letter language code corresponding to ``aug_toks``.
            Used to load a weighted distribution of language-appropriate characters
            that are randomly selected for substitution. More common characters
            are more likely to be substituted. If not specified, ascii letters and
            digits are randomly selected with equal probability.

    Returns:
        New, augmented sequence of tokens.

    Note:
        This transform requires :class:`textacy.datasets.UDHR` to be downloaded
        to work properly, since this is the data source for character weights when
        deciding which char(s) to insert.
    """
    _validate_aug_toks(aug_toks)
    char_weights = aug_utils.get_char_weights(lang or "xx")
    cand_idxs = [
        idx for idx, aug_tok in enumerate(aug_toks)
        if aug_tok.is_word and len(aug_tok.text) >= 3
    ]
    rand_idxs = set(_select_random_candidates(cand_idxs, num))
    if not rand_idxs:
        return aug_toks[:]

    rand_chars = iter(
        random.choices(
            [char for char, _ in char_weights],
            weights=[weight for _, weight in char_weights],
            k=len(rand_idxs),
        )
    )
    new_aug_toks = []
    for idx, aug_tok in enumerate(aug_toks):
        if idx in rand_idxs:
            text_list = list(aug_tok.text)
            rand_char_idx = random.choice(range(len(text_list)))
            text_list[rand_char_idx] = next(rand_chars)
            new_aug_toks.append(
                aug_utils.AugTok(
                    text="".join(text_list),
                    ws=aug_tok.ws,
                    pos=aug_tok.pos,
                    is_word=aug_tok.is_word,
                    syns=aug_tok.syns,
                )
            )
        else:
            new_aug_toks.append(aug_tok)
    return new_aug_toks


def insert_chars(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
    lang: Optional[str] = None,
) -> List[aug_utils.AugTok]:
    """
    Randomly insert a character into randomly-selected words,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through character insertion.
        num: If int, maximum number of words to modify with a random character insertion;
            if float, probability that a given word will be modified.
        lang: Standard, two-letter language code corresponding to ``aug_toks``.
            Used to load a weighted distribution of language-appropriate characters
            that are randomly selected for substitution. More common characters
            are more likely to be substituted. If not specified, ascii letters and
            digits are randomly selected with equal probability.

    Returns:
        New, augmented sequence of tokens.

    Note:
        This transform requires :class:`textacy.datasets.UDHR` to be downloaded
        to work properly, since this is the data source for character weights when
        deciding which char(s) to insert.
    """
    _validate_aug_toks(aug_toks)
    char_weights = aug_utils.get_char_weights(lang or "xx")
    cand_idxs = [
        idx for idx, aug_tok in enumerate(aug_toks)
        if aug_tok.is_word and len(aug_tok.text) >= 3
    ]
    rand_idxs = set(_select_random_candidates(cand_idxs, num))
    if not rand_idxs:
        return aug_toks[:]

    rand_chars = iter(
        random.choices(
            [char for char, _ in char_weights],
            weights=[weight for _, weight in char_weights],
            k=len(rand_idxs),
        )
    )
    new_aug_toks = []
    for idx, aug_tok in enumerate(aug_toks):
        if idx in rand_idxs:
            text_list = list(aug_tok.text)
            rand_char_idx = random.choice(range(len(text_list)))
            text_list.insert(rand_char_idx, next(rand_chars))
            new_aug_toks.append(
                aug_utils.AugTok(
                    text="".join(text_list),
                    ws=aug_tok.ws,
                    pos=aug_tok.pos,
                    is_word=aug_tok.is_word,
                    syns=aug_tok.syns,
                )
            )
        else:
            new_aug_toks.append(aug_tok)
    return new_aug_toks


def swap_chars(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
) -> List[aug_utils.AugTok]:
    """
    Randomly swap two *adjacent* characters in randomly-selected words,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through character swapping.
        num: If int, maximum number of words to modify with a random character swap;
            if float, probability that a given word will be modified.

    Returns:
        New, augmented sequence of tokens.
    """
    _validate_aug_toks(aug_toks)
    cand_idxs = [
        idx for idx, aug_tok in enumerate(aug_toks)
        if aug_tok.is_word and len(aug_tok.text) >= 3
    ]
    rand_idxs = set(_select_random_candidates(cand_idxs, num))
    if not rand_idxs:
        return aug_toks[:]

    new_aug_toks = []
    for idx, aug_tok in enumerate(aug_toks):
        if idx in rand_idxs:
            text_list = list(aug_tok.text)
            idx = random.choice(range(1, len(text_list)))
            text_list[idx - 1], text_list[idx] = text_list[idx], text_list[idx - 1]
            new_aug_toks.append(
                aug_utils.AugTok(
                    text="".join(text_list),
                    ws=aug_tok.ws,
                    pos=aug_tok.pos,
                    is_word=aug_tok.is_word,
                    syns=aug_tok.syns,
                )
            )
        else:
            new_aug_toks.append(aug_tok)
    return new_aug_toks


def delete_chars(
    aug_toks: List[aug_utils.AugTok],
    *,
    num: Union[int, float] = 1,
) -> List[aug_utils.AugTok]:
    """
    Randomly delete a character in randomly-selected words,
    up to ``num`` times or with a probability of ``num``.

    Args:
        aug_toks: Sequence of tokens to augment through character deletion.
        num: If int, maximum number of words to modify with a random character deletion;
            if float, probability that a given word will be modified.

    Returns:
        New, augmented sequence of tokens.
    """
    _validate_aug_toks(aug_toks)
    cand_idxs = [
        idx for idx, aug_tok in enumerate(aug_toks)
        if aug_tok.is_word and len(aug_tok.text) >= 3
    ]
    rand_idxs = set(_select_random_candidates(cand_idxs, num))
    if not rand_idxs:
        return aug_toks[:]

    new_aug_toks = []
    for idx, aug_tok in enumerate(aug_toks):
        if idx in rand_idxs:
            rand_char_idx = random.choice(range(len(aug_tok.text)))
            text = "".join(
                char for char_idx, char in enumerate(aug_tok.text)
                if char_idx != rand_char_idx
            )
            new_aug_toks.append(
                aug_utils.AugTok(
                    text=text,
                    ws=aug_tok.ws,
                    pos=aug_tok.pos,
                    is_word=aug_tok.is_word,
                    syns=aug_tok.syns,
                )
            )
        else:
            new_aug_toks.append(aug_tok)
    return new_aug_toks


def _validate_aug_toks(aug_toks):
    if not (isinstance(aug_toks, list) and isinstance(aug_toks[0], aug_utils.AugTok)):
        raise TypeError(
            "aug_toks must be of type List[:obj:`AugTok`], not {}[{}]".format(
                type(aug_toks), type(aug_toks[0]))
        )


def _select_random_candidates(cands, num):
    """
    Args:
        cands (List[obj])
        num (int or float)

    Returns:
        List[obj]
    """
    if isinstance(num, int) and num >= 0:
        rand_cands = random.sample(cands, min(num, len(cands)))
    elif isinstance(num, float) and 0.0 <= num <= 1.0:
        rand_cands = [cand for cand in cands if random.random() < num]
    else:
        raise ValueError(
            "num={} is invalid; must be an int >= 0 or a float in [0.0, 1.0]".format(num)
        )
    return rand_cands



"""
Text Statistics
---------------

Compute a variety of basic counts and readability statistics for documents.
"""
import functools
import logging
from math import sqrt
from typing import Dict, Optional

from cachetools import cached
from cachetools.keys import hashkey
from cytoolz import itertoolz

from . import cache, extract


LOGGER = logging.getLogger(__name__)


class TextStats:
    """
    Compute a variety of basic counts and readability statistics for a given
    document. For example:

    .. code-block:: pycon

        >>> text = next(textacy.datasets.CapitolWords().texts(limit=1))
        >>> doc = textacy.make_spacy_doc(text)
        >>> ts = TextStats(doc)
        >>> ts.n_words
        136
        >>> ts.flesch_kincaid_grade_level
        11.817647058823532
        >>> ts.basic_counts
        {'n_chars': 685,
         'n_long_words': 43,
         'n_monosyllable_words': 90,
         'n_polysyllable_words': 24,
         'n_sents': 6,
         'n_syllables': 214,
         'n_unique_words': 80,
         'n_words': 136}
        >>> ts.readability_stats
        {'automated_readability_index': 13.626495098039214,
         'coleman_liau_index': 12.509300816176474,
         'flesch_kincaid_grade_level': 11.817647058823532,
         'flesch_reading_ease': 50.707745098039254,
         'gulpease_index': 51.86764705882353,
         'gunning_fog_index': 16.12549019607843,
         'lix': 54.28431372549019,
         'smog_index': 14.554592549557764,
         'wiener_sachtextformel': 8.266410784313727}

    Args:
        doc (:class:`spacy.tokens.Doc`): A text document processed
            by spacy. Need only be tokenized.

    Attributes:
        n_sents (int): Number of sentences in ``doc``.
        n_words (int): Number of words in ``doc``, including numbers + stop words
            but excluding punctuation.
        n_chars (int): Number of characters for all words in ``doc``.
        n_syllables (int): Number of syllables for all words in ``doc``.
        n_unique_words (int): Number of unique (lower-cased) words in ``doc``.
        n_long_words (int): Number of words in ``doc`` with 7 or more characters.
        n_monosyllable_words (int): Number of words in ``doc`` with 1 syllable only.
        n_polysyllable_words (int): Number of words in ``doc`` with 3 or more syllables.
            Note: Since this excludes words with exactly 2 syllables, it's likely
            that ``n_monosyllable_words + n_polysyllable_words != n_words``.
        flesch_kincaid_grade_level (float): see :func:`flesch_kincaid_grade_level()`
        flesch_reading_ease (float): see :func:`flesch_reading_ease()`
        smog_index (float): see :func:`smog_index()`
        gunning_fog_index (float): see :func:`gunning_fog_index()`
        coleman_liau_index (float): see :func:`coleman_liau_index()`
        automated_readability_index (float): see :func:`automated_readability_index()`
        lix (float): see :func:`lix()`
        gulpease_index (float): see :func:`gulpease_index()`
        wiener_sachtextformel (float): see :func:`wiener_sachtextformel()`
            Note: This always returns variant #1.
        basic_counts (Dict[str, int]): Mapping of basic count names to values,
            where basic counts are the attributes listed above between ``n_sents``
            and ``n_polysyllable_words``.
        readability_stats (Dict[str, float]): Mapping of readability statistic
            names to values, where readability stats are the attributes listed
            above between ``flesch_kincaid_grade_level`` and ``wiener_sachtextformel``.

    Raises:
        ValueError: If ``doc`` is not a :class:`spacy.tokens.Doc`.
    """

    def __init__(self, doc):
        self.lang = doc.vocab.lang
        self.n_sents = itertoolz.count(doc.sents) if doc.is_sentenced else None
        # get objs for basic count computations
        hyphenator = load_hyphenator(lang=self.lang)
        words = tuple(
            extract.words(doc, filter_punct=True, filter_stops=False, filter_nums=False)
        )
        syllables_per_word = tuple(
            len(hyphenator.positions(word.lower_)) + 1 for word in words
        )
        chars_per_word = tuple(len(word) for word in words)
        # compute basic counts needed for most readability stats
        self.n_words = len(words)
        self.n_unique_words = len({word.lower for word in words})
        self.n_chars = sum(chars_per_word)
        self.n_long_words = sum(1 for cpw in chars_per_word if cpw >= 7)
        self.n_syllables = sum(syllables_per_word)
        self.n_monosyllable_words = sum(1 for spw in syllables_per_word if spw == 1)
        self.n_polysyllable_words = sum(1 for spw in syllables_per_word if spw >= 3)

    @property
    def flesch_kincaid_grade_level(self):
        return flesch_kincaid_grade_level(self.n_syllables, self.n_words, self.n_sents)

    @property
    def flesch_reading_ease(self):
        return flesch_reading_ease(
            self.n_syllables, self.n_words, self.n_sents, lang=self.lang
        )

    @property
    def smog_index(self):
        return smog_index(self.n_polysyllable_words, self.n_sents)

    @property
    def gunning_fog_index(self):
        return gunning_fog_index(self.n_words, self.n_polysyllable_words, self.n_sents)

    @property
    def coleman_liau_index(self):
        return coleman_liau_index(self.n_chars, self.n_words, self.n_sents)

    @property
    def automated_readability_index(self):
        return automated_readability_index(self.n_chars, self.n_words, self.n_sents)

    @property
    def lix(self):
        return lix(self.n_words, self.n_long_words, self.n_sents)

    @property
    def gulpease_index(self):
        return gulpease_index(self.n_chars, self.n_words, self.n_sents)

    @property
    def wiener_sachtextformel(self):
        return wiener_sachtextformel(
            self.n_words,
            self.n_polysyllable_words,
            self.n_monosyllable_words,
            self.n_long_words,
            self.n_sents,
            variant=1,
        )

    @property
    def basic_counts(self) -> Dict[str, int]:
        return {
            "n_sents": self.n_sents,
            "n_words": self.n_words,
            "n_chars": self.n_chars,
            "n_syllables": self.n_syllables,
            "n_unique_words": self.n_unique_words,
            "n_long_words": self.n_long_words,
            "n_monosyllable_words": self.n_monosyllable_words,
            "n_polysyllable_words": self.n_polysyllable_words,
        }

    @property
    def readability_stats(self) -> Optional[Dict[str, float]]:
        if self.n_words == 0:
            LOGGER.warning(
                "readability stats can't be computed because doc has 0 words"
            )
            return None
        elif self.n_sents is None:
            LOGGER.warning(
                "readability stats can't be computed because doc has not been "
                "segmented into sentences"
            )
            return None
        return {
            "flesch_kincaid_grade_level": self.flesch_kincaid_grade_level,
            "flesch_reading_ease": self.flesch_reading_ease,
            "smog_index": self.smog_index,
            "gunning_fog_index": self.gunning_fog_index,
            "coleman_liau_index": self.coleman_liau_index,
            "automated_readability_index": self.automated_readability_index,
            "lix": self.lix,
            "gulpease_index": self.gulpease_index,
            "wiener_sachtextformel": self.wiener_sachtextformel,
        }


def flesch_kincaid_grade_level(n_syllables: int, n_words: int, n_sents: int) -> float:
    """
    Readability score used widely in education, whose value estimates the U.S.
    grade level / number of years of education required to understand a text.
    Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Flesch%E2%80%93Kincaid_readability_tests#Flesch.E2.80.93Kincaid_grade_level
    """
    return (11.8 * n_syllables / n_words) + (0.39 * n_words / n_sents) - 15.59


def flesch_reading_ease(
    n_syllables: int,
    n_words: int,
    n_sents: int,
    *,
    lang: Optional[str] = None,
) -> float:
    """
    Readability score usually in the range [0, 100], related (inversely) to
    :func:`flesch_kincaid_grade_level()`. Higher value => easier text.

    Note:
        Constant weights in this formula are language-dependent;
        if ``lang`` is null, the English-language formulation is used.

    References:
        English: https://en.wikipedia.org/wiki/Flesch%E2%80%93Kincaid_readability_tests#Flesch_reading_ease
        German: https://de.wikipedia.org/wiki/Lesbarkeitsindex#Flesch-Reading-Ease
        Spanish: ?
        French: ?
        Italian: https://it.wikipedia.org/wiki/Formula_di_Flesch
        Dutch: ?
        Portuguese: https://pt.wikipedia.org/wiki/Legibilidade_de_Flesch
        Russian: https://ru.wikipedia.org/wiki/%D0%98%D0%BD%D0%B4%D0%B5%D0%BA%D1%81_%D1%83%D0%B4%D0%BE%D0%B1%D0%BE%D1%87%D0%B8%D1%82%D0%B0%D0%B5%D0%BC%D0%BE%D1%81%D1%82%D0%B8
    """
    if lang is None or lang == "en":
        return 206.835 - (1.015 * n_words / n_sents) - (84.6 * n_syllables / n_words)
    elif lang == "de":
        return 180.0 - (n_words / n_sents) - (58.5 * n_syllables / n_words)
    elif lang == "es":
        return 206.84 - (1.02 * n_words / n_sents) - (60.0 * n_syllables / n_words)
    elif lang == "fr":
        return 207.0 - (1.015 * n_words / n_sents) - (73.6 * n_syllables / n_words)
    elif lang == "it":
        return 217.0 - (1.3 * n_words / n_sents) - (60.0 * n_syllables / n_words)
    elif lang == "nl":
        return 206.84 - (0.93 * n_words / n_sents) - (77.0 * n_syllables / n_words)
    elif lang == "pt":
        return 248.835 - (1.015 * n_words / n_sents) - (84.6 * n_syllables / n_words)
    elif lang == "ru":
        return 206.835 - (1.3 * n_words / n_sents) - (60.1 * n_syllables / n_words)
    else:
        langs = ["en", "de", "es", "fr", "it", "nl", "pt", "ru"]
        raise ValueError(
            "Flesch Reading Ease is only implemented for these languages: {}. "
            'Passing `lang=None` falls back to "en" (English)'.format(langs)
        )


def smog_index(n_polysyllable_words: int, n_sents: int) -> float:
    """
    Readability score commonly used in healthcare, whose value estimates the
    number of years of education required to understand a text, similar to
    :func:`flesch_kincaid_grade_level()` and intended as a substitute for
    :func:`gunning_fog_index()`. Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/SMOG
    """
    if n_sents < 30:
        LOGGER.warning("SMOG score may be unreliable for n_sents < 30")
    return (1.0430 * sqrt(30 * n_polysyllable_words / n_sents)) + 3.1291


def gunning_fog_index(n_words: int, n_polysyllable_words: int, n_sents: int) -> float:
    """
    Readability score whose value estimates the number of years of education
    required to understand a text, similar to :func:`flesch_kincaid_grade_level()`
    and :func:`smog_index()`. Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Gunning_fog_index
    """
    return 0.4 * ((n_words / n_sents) + (100 * n_polysyllable_words / n_words))


def coleman_liau_index(n_chars: int, n_words: int, n_sents: int) -> float:
    """
    Readability score whose value estimates the number of years of education
    required to understand a text, similar to :func:`flesch_kincaid_grade_level()`
    and :func:`smog_index()`, but using characters instead of syllables.
    Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Coleman%E2%80%93Liau_index"""
    return (5.879851 * n_chars / n_words) - (29.587280 * n_sents / n_words) - 15.800804


def automated_readability_index(n_chars: int, n_words: int, n_sents: int) -> float:
    """
    Readability score whose value estimates the U.S. grade level required to
    understand a text, most similarly to :func:`flesch_kincaid_grade_level()`,
    but using characters instead of syllables like :func:`coleman_liau_index()`.
    Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/Automated_readability_index
    """
    return (4.71 * n_chars / n_words) + (0.5 * n_words / n_sents) - 21.43


def lix(n_words: int, n_long_words: int, n_sents: int) -> float:
    """
    Readability score commonly used in Sweden, whose value estimates the
    difficulty of reading a foreign text. Higher value => more difficult text.

    References:
        https://en.wikipedia.org/wiki/LIX
    """
    return (n_words / n_sents) + (100 * n_long_words / n_words)


def wiener_sachtextformel(
    n_words: int,
    n_polysyllable_words: int,
    n_monosyllable_words: int,
    n_long_words: int,
    n_sents: int,
    *,
    variant: int = 1,
) -> float:
    """
    Readability score for German-language texts, whose value estimates the grade
    level required to understand a text. Higher value => more difficult text.

    References:
        https://de.wikipedia.org/wiki/Lesbarkeitsindex#Wiener_Sachtextformel
    """
    ms = 100 * n_polysyllable_words / n_words
    sl = n_words / n_sents
    iw = 100 * n_long_words / n_words
    es = 100 * n_monosyllable_words / n_words
    if variant == 1:
        return (0.1935 * ms) + (0.1672 * sl) + (0.1297 * iw) - (0.0327 * es) - 0.875
    elif variant == 2:
        return (0.2007 * ms) + (0.1682 * sl) + (0.1373 * iw) - 2.779
    elif variant == 3:
        return (0.2963 * ms) + (0.1905 * sl) - 1.1144
    elif variant == 4:
        return (0.2744 * ms) + (0.2656 * sl) - 1.693
    else:
        raise ValueError("``variant`` value invalid; must be 1, 2, 3, or 4")


def gulpease_index(n_chars: int, n_words: int, n_sents: int) -> float:
    """
    Readability score for Italian-language texts, whose value is in the range
    [0, 100] similar to :func:`flesch_reading_ease()`. Higher value =>
    easier text.

    References:
        https://it.wikipedia.org/wiki/Indice_Gulpease
    """
    return (300 * n_sents / n_words) - (10 * n_chars / n_words) + 89


@cached(cache.LRU_CACHE, key=functools.partial(hashkey, "hyphenator"))
def load_hyphenator(lang: str):
    """
    Load an object that hyphenates words at valid points, as used in LaTex typesetting.

    Args:
        lang (str): Standard 2-letter language abbreviation. To get a list of
            valid values::

                >>> import pyphen; pyphen.LANGUAGES

    Returns:
        :class:`pyphen.Pyphen()`

    Note:
        While hyphenation points always fall on syllable divisions,
        not all syllable divisions are valid hyphenation points. But it's decent.
    """
    import pyphen

    LOGGER.debug('Loading "%s" language hyphenator', lang)
    return pyphen.Pyphen(lang=lang)



"""
Dataset Utils
-------------

Shared functionality for downloading, naming, and extracting the contents
of datasets, as well as filtering for particular subsets.
"""
import logging
import os
import shutil
import tarfile
import urllib.parse
import zipfile

from .. import constants, utils
from ..io import utils as io_utils

LOGGER = logging.getLogger(__name__)


def download_file(url, *, filename=None, dirpath=constants.DEFAULT_DATA_DIR, force=False):
    utils.deprecated(
        "This function has been moved to `textacy.io.utils.download_file()` "
        "and is aliased here only for backwards compatibility. "
        "This alias will be removed in v0.10.0.",
        action="once",
    )
    return io_utils.download_file(url, filename=filename, dirpath=dirpath, force=force)


def get_filename_from_url(url):
    utils.deprecated(
        "This function has been moved to `textacy.io.utils.get_filename_from_url()` "
        "and is aliased here only for backwards compatibility. "
        "This alias will be removed in v0.10.0.",
        action="once",
    )
    return io_utils.get_filename_from_url(url)


def unpack_archive(filepath, *, extract_dir=None):
    utils.deprecated(
        "This function has been moved to `textacy.io.utils.unpack_archive()` "
        "and is aliased here only for backwards compatibility. "
        "This alias will be removed in v0.10.0.",
        action="once",
    )
    return io_utils.unpack_archive(filepath, extract_dir=extract_dir)


def validate_set_member_filter(filter_vals, vals_type, valid_vals=None):
    utils.deprecated(
        "This function has been moved to `textacy.utils.validate_set_members()` "
        "and is aliased here only for backwards compatibility. "
        "This alias will be removed in v0.10.0.",
        action="once",
    )
    return utils.validate_set_members(filter_vals, vals_type, valid_vals=valid_vals)


def validate_and_clip_range_filter(filter_range, full_range, val_type=None):
    utils.deprecated(
        "This function has been moved to `textacy.utils.validate_and_clip_range()` "
        "and is aliased here only for backwards compatibility. "
        "This alias will be removed in v0.10.0.",
        action="once",
    )
    return utils.validate_and_clip_range(filter_range, full_range, val_type=val_type)



"""
ConceptNet
----------

ConceptNet is a multilingual knowledge base, representing common words and phrases
and the common-sense relationships between them. This information is collected from
a variety of sources, including crowd-sourced resources (e.g. Wiktionary, Open Mind
Common Sense), games with a purpose (e.g. Verbosity, nadya.jp), and expert-created
resources (e.g. WordNet, JMDict).

The interface in textacy gives access to several key relationships between terms
that are useful in a variety of NLP tasks:

    - antonyms: terms that are opposites of each other in some relevant way
    - hyponyms: terms that are subtypes or specific instances of other terms
    - meronyms: terms that are parts of other terms
    - synonyms: terms that are sufficiently similar that they may be used interchangeably
"""
import collections
import json
import logging

from spacy.tokens import Span, Token
from tqdm import tqdm

from .. import constants
from .. import io as tio
from .. import utils
from .base import Resource


LOGGER = logging.getLogger(__name__)

NAME = "concept_net"
META = {
    "site_url": "http://conceptnet.io",
    "publication_url": "https://arxiv.org/abs/1612.03975",
    "description": (
        "An open, multilingual semantic network of general knowledge, "
        "designed to help computers understand the meanings of words."
    )
}
DOWNLOAD_ROOT = "https://s3.amazonaws.com/conceptnet/downloads/{year}/edges/conceptnet-assertions-{version}.csv.gz"


class ConceptNet(Resource):
    """
    Interface to ConceptNet, a multilingual knowledge base representing common words
    and phrases and the common-sense relationships between them.

    Download the data (one time only!), and save its contents to disk::

        >>> import textacy.resources
        >>> rs = textacy.resources.ConceptNet()
        >>> rs.download()
        >>> rs.info
        {'name': 'concept_net',
         'site_url': 'http://conceptnet.io',
         'publication_url': 'https://arxiv.org/abs/1612.03975',
         'description': 'An open, multilingual semantic network of general knowledge, designed to help computers understand the meanings of words.'}

    Access other same-language terms related to a given term in a variety of ways::

        >>> rs.get_synonyms("spouse", lang="en", sense="n")
        ['mate', 'married person', 'better half', 'partner']
        >>> rs.get_antonyms("love", lang="en", sense="v")
        ['detest', 'hate', 'loathe']
        >>> rs.get_hyponyms("marriage", lang="en", sense="n")
        ['cohabitation situation', 'union', 'legal agreement', 'ritual', 'family', 'marital status']

    **Note:** The very first time a given relationship is accessed, the full ConceptNet db
    must be parsed and split for fast future access. This can take a couple minutes;
    be patient.

    When passing a spaCy ``Token`` or ``Span``, the corresponding ``lang`` and ``sense``
    are inferred automatically from the object::

        >>> text = "The quick brown fox jumps over the lazy dog."
        >>> doc = textacy.make_spacy_doc(text, lang="en")
        >>> rs.get_synonyms(doc[1])  # quick
        ['flying', 'fast', 'rapid', 'ready', 'straightaway', 'nimble', 'speedy', 'warm']
        >>> rs.get_synonyms(doc[4:5])  # jumps over
        ['leap', 'startle', 'hump', 'flinch', 'jump off', 'skydive', 'jumpstart', ...]

    Many terms won't have entries, for actual linguistic reasons or because the db's
    coverage of a given language's vocabulary isn't comprehensive::

        >>> rs.get_meronyms(doc[3])  # fox
        []
        >>> rs.get_antonyms(doc[7])  # lazy
        []

    Args:
        data_dir (str or :class:`pathlib.Path`): Path to directory on disk
            under which resource data is stored, i.e. ``/path/to/data_dir/concept_net``.
        version ({"5.7.0", "5.6.0", "5.5.5"}): Version string of the ConceptNet db
            to use. Since newer versions typically represent improvements over earlier
            versions, you'll probably want "5.7.0" (the default value).
    """

    _version_years = {"5.7.0": 2019, "5.6.0": 2018, "5.5.5": 2017}
    _pos_map = {"NOUN": "n", "VERB": "v", "ADJ": "a", "ADV": "r"}

    def __init__(
        self,
        data_dir=constants.DEFAULT_DATA_DIR.joinpath(NAME),
        version="5.7.0",
    ):
        super().__init__(NAME, meta=META)
        self.version = version
        self.data_dir = utils.to_path(data_dir).resolve().joinpath(self.version)
        self._filename = "conceptnet-assertions-{}.csv.gz".format(self.version)
        self._filepath = self.data_dir.joinpath(self._filename)
        self._antonyms = None
        self._hyponyms = None
        self._meronyms = None
        self._synonyms = None

    def download(self, *, force=False):
        """
        Download resource data as a gzipped csv file,
        then save it to disk under the :attr:`ConceptNet.data_dir` directory.

        Args:
            force (bool): If True, download resource data, even if it already
                exists on disk; otherwise, don't re-download the data.
        """
        url = DOWNLOAD_ROOT.format(
            version=self.version, year=self._version_years[self.version])
        tio.download_file(
            url,
            filename=self._filename,
            dirpath=self.data_dir,
            force=force,
        )

    @property
    def filepath(self):
        """
        str: Full path on disk for the ConceptNet gzipped csv file
        corresponding to the given :attr:`ConceptNet.data_dir`.
        """
        if self._filepath.is_file():
            return str(self._filepath)
        else:
            return None

    def _get_relation_data(self, relation, is_symmetric=False):
        if not self.filepath:
            raise OSError(
                "resource file {} not found;\n"
                "has the data been downloaded yet?".format(self._filepath)
            )
        rel_fname = "{}.json.gz".format(_split_uri(relation)[1].lower())
        rel_fpath = self.data_dir.joinpath(rel_fname)
        if rel_fpath.is_file():
            LOGGER.debug("loading data for '%s' relation from %s", relation, rel_fpath)
            return next(
                tio.read_json(rel_fpath, mode="rt", encoding="utf-8", lines=False)
            )
        else:
            rel_data = collections.defaultdict(
                lambda: collections.defaultdict(
                    lambda: collections.defaultdict(set)
                )
            )
            LOGGER.info(
                "preparing data for '%s' relation; this may take a while...", relation)
            rows = tio.read_csv(self.filepath, delimiter="\t", quoting=1)
            with tqdm() as pbar:
                for row in rows:
                    pbar.update(1)
                    _, rel_type, start_uri, end_uri, _ = row
                    if rel_type < relation:
                        continue
                    elif rel_type > relation:
                        break
                    start_lang, start_term, start_sense = _parse_concept_uri(start_uri)
                    end_lang, end_term, end_sense = _parse_concept_uri(end_uri)
                    if start_lang == end_lang and start_term != end_term:
                        rel_data[start_lang][start_term][start_sense].add(end_term)
                        if is_symmetric:
                            rel_data[start_lang][end_term][end_sense].add(start_term)
            # make relation data json-able (i.e. cast set => list)
            for terms in rel_data.values():
                for senses in terms.values():
                    for sense, rel_terms in senses.items():
                        senses[sense] = list(rel_terms)
            LOGGER.info("saving data for '%s' relation to %s", relation, rel_fpath)
            tio.write_json(rel_data, rel_fpath, mode="wt", encoding="utf-8")
            return rel_data

    def _get_relation_values(self, rel_data, term, lang=None, sense=None):
        """
        Args:
            term (str or :class:`spacy.tokens.Token` or :class:`spacy.tokens.Span`)
            lang (str)
            sense (str)

        Returns:
            List[str]
        """
        if lang is not None and lang not in rel_data:
            raise ValueError(
                "lang='{}' is invalid; valid langs are {}".format(
                    lang, sorted(rel_data.keys())
                )
            )
        if sense is not None:
            # let's be kind and automatically convert standard POS strings
            if sense in self._pos_map.keys():
                sense = self._pos_map[sense]
            # otherwise, return no results as is done when auto-inferrence of sense
            # results in an invalid value
            elif sense not in self._pos_map.values():
                return []
        if isinstance(term, str):
            if not (lang and sense):
                raise ValueError(
                    "if `term` is a string, both `lang` and `sense` must be specified")
            else:
                norm_terms = [term.replace(" ", "_").lower()]
        elif isinstance(term, (Span, Token)):
            norm_terms = [
                term.text.replace(" ", "_").lower(),
                term.lemma_.replace(" ", "_").lower(),
            ]
            if not lang:
                try:
                    lang = term.lang_  # token
                except AttributeError:
                    lang = term[0].lang_  # span
            if not sense:
                try:
                    sense = self._pos_map[term.pos_]  # token
                except AttributeError:
                    sense = self._pos_map[term[0].pos_]  # span
                except KeyError:
                    return []
        else:
            raise TypeError(
                "`term` must be one of {}, not {}".format(
                    {str, Span, Token}, type(term)
                )
            )
        # TODO: implement an out-of-vocabulary strategy? for example,
        # https://github.com/commonsense/conceptnet-numberbatch#out-of-vocabulary-strategy
        for norm_term in norm_terms:
            try:
                return rel_data[lang][norm_term].get(sense, [])
            except KeyError:
                pass
        return []

    @property
    def antonyms(self):
        """
        Dict[str, Dict[str, Dict[str, List[str]]]]: Mapping of language code to term to
        sense to set of term's antonyms -- opposites of the term in some relevant way,
        like being at opposite ends of a scale or fundamentally similar but with
        a key difference between them -- such as black <=> white or hot <=> cold. Note
        that this relationship is symmetric.

        Based on the "/r/Antonym" relation in ConceptNet.
        """
        if not self._antonyms:
            self._antonyms = self._get_relation_data("/r/Antonym", is_symmetric=True)
        return self._antonyms

    def get_antonyms(self, term, *, lang=None, sense=None):
        """
        Args:
            term (str or :class:`spacy.tokens.Token` or :class:`spacy.tokens.Span`)
            lang (str): Standard code for the language of ``term``.
            sense (str): Sense in which ``term`` is used in context, which in practice
                is just its part of speech. Valid values: "n" or "NOUN", "v" or "VERB",
                "a" or "ADJ", "r" or "ADV".

        Returns:
            List[str]
        """
        return self._get_relation_values(self.antonyms, term, lang=lang, sense=sense)

    @property
    def hyponyms(self):
        """
        Dict[str, Dict[str, Dict[str, List[str]]]]: Mapping of language code to term to
        sense to set of term's hyponyms -- subtypes or specific instances of the term --
        such as car => vehicle or Chicago => city. Every A is a B.

        Based on the "/r/IsA" relation in ConceptNet.
        """
        if not self._hyponyms:
            self._hyponyms = self._get_relation_data("/r/IsA", is_symmetric=False)
        return self._hyponyms

    def get_hyponyms(self, term, *, lang=None, sense=None):
        """
        Args:
            term (str or :class:`spacy.tokens.Token` or :class:`spacy.tokens.Span`)
            lang (str): Standard code for the language of ``term``.
            sense (str): Sense in which ``term`` is used in context, which in practice
                is just its part of speech. Valid values: "n" or "NOUN", "v" or "VERB",
                "a" or "ADJ", "r" or "ADV".

        Returns:
            List[str]
        """
        return self._get_relation_values(self.hyponyms, term, lang=lang, sense=sense)

    @property
    def meronyms(self):
        """
        Dict[str, Dict[str, Dict[str, List[str]]]]: Mapping of language code to term to
        sense to set of term's meronyms -- parts of the term -- such as gearshift => car.

        Based on the "/r/PartOf" relation in ConceptNet.
        """
        if not self._meronyms:
            self._meronyms = self._get_relation_data("/r/PartOf", is_symmetric=False)
        return self._meronyms

    def get_meronyms(self, term, *, lang=None, sense=None):
        """
        Args:
            term (str or :class:`spacy.tokens.Token` or :class:`spacy.tokens.Span`)
            lang (str): Standard code for the language of ``term``.
            sense (str): Sense in which ``term`` is used in context, which in practice
                is just its part of speech. Valid values: "n" or "NOUN", "v" or "VERB",
                "a" or "ADJ", "r" or "ADV".

        Returns:
            List[str]
        """
        return self._get_relation_values(self.meronyms, term, lang=lang, sense=sense)

    @property
    def synonyms(self):
        """
        Dict[str, Dict[str, Dict[str, List[str]]]]: Mapping of language code to term to
        sense to set of term's synonyms -- sufficiently similar concepts that they may
        be used interchangeably -- such as sunlight <=> sunshine. Note that
        this relationship is symmetric.

        Based on the "/r/Synonym" relation in ConceptNet.
        """
        if not self._synonyms:
            self._synonyms = self._get_relation_data("/r/Synonym", is_symmetric=True)
        return self._synonyms

    def get_synonyms(self, term, *, lang=None, sense=None):
        """
        Args:
            term (str or :class:`spacy.tokens.Token` or :class:`spacy.tokens.Span`)
            lang (str): Standard code for the language of ``term``.
            sense (str): Sense in which ``term`` is used in context, which in practice
                is just its part of speech. Valid values: "n" or "NOUN", "v" or "VERB",
                "a" or "ADJ", "r" or "ADV".

        Returns:
            List[str]
        """
        return self._get_relation_values(self.synonyms, term, lang=lang, sense=sense)


def _split_uri(uri):
    """
    Get slash-delimited parts of a ConceptNet URI.

    Args:
        uri (str)

    Returns:
        List[str]
    """
    uri = uri.lstrip("/")
    if not uri:
        return []
    return uri.split("/")


def _parse_concept_uri(uri):
    """
    Extract language, term, and sense from a ConceptNet "concept" URI.

    Args:
        uri (str)

    Returns:
        Tuple[str, str, str]: Language, term, sense.
    """
    if not uri.startswith("/c/"):
        raise ValueError("invalid concept uri: {}".format(uri))
    uri = _split_uri(uri)
    if len(uri) == 3:
        _, lang, term = uri
        sense = None
    elif len(uri) == 4:
        _, lang, term, sense = uri
    elif len(uri) > 4:
        _, lang, term, sense, *_ = uri
    elif len(uri) < 3:
        raise ValueError("not enough parts in uri: {}".format(uri))
    term = term.replace("_", " ")
    return lang, term, sense



import math

import networkx as nx

try:
    import matplotlib.pyplot as plt
except ImportError:
    pass


RC_PARAMS = {
    "axes.axisbelow": True,
    "axes.edgecolor": ".8",
    "axes.facecolor": "white",
    "axes.grid": False,
    "axes.labelcolor": ".15",
    "axes.linewidth": 1.0,
    "figure.facecolor": "white",
    "font.family": ["sans-serif"],
    "font.sans-serif": ["Arial", "Liberation Sans", "sans-serif"],
    "grid.color": ".8",
    "grid.linestyle": "-",
    "image.cmap": "Greys",
    "legend.frameon": False,
    "legend.numpoints": 1,
    "legend.scatterpoints": 1,
    "lines.solid_capstyle": "round",
    "text.color": ".15",
    "xtick.color": ".15",
    "xtick.direction": "out",
    "xtick.major.size": 0.0,
    "xtick.minor.size": 0.0,
    "ytick.color": ".15",
    "ytick.direction": "out",
    "ytick.major.size": 0.0,
    "ytick.minor.size": 0.0,
}


def draw_semantic_network(
    graph,
    *,
    node_weights=None,
    spread=3.0,
    draw_nodes=False,
    base_node_size=300,
    node_alpha=0.25,
    line_width=0.5,
    line_alpha=0.1,
    base_font_size=12,
    save=False,
):
    """
    Draw a semantic network with nodes representing either terms or sentences,
    edges representing coocurrence or similarity, and positions given by a force-
    directed layout.

    Args:
        graph (``networkx.Graph``):
        node_weights (dict): mapping of node: weight, used to size node labels
            (and, optionally, node circles) according to their weight
        spread (float): number that drives the spread of the network; higher
            values give more spread-out networks
        draw_nodes (bool): if True, circles are drawn under the node labels
        base_node_size (int): if `node_weights` not given and `draw_nodes` is True,
            this is the size of all nodes in the network; if `node_weights` _is_
            given, node sizes will be scaled against this value based on their
            weights compared to the max weight
        node_alpha (float): alpha of the circular nodes drawn behind labels if
            `draw_nodes` is True
        line_width (float): width of the lines (edges) drawn between nodes
        line_alpha (float): alpha of the lines (edges) drawn between nodes
        base_font_size (int): if `node_weights` not given, this is the font size
            used to draw all labels; otherwise, font sizes will be scaled against
            this value based on the corresponding node weights compared to the max
        save (str): give the full /path/to/fname on disk to save figure (optional)

    Returns:
        :obj:`matplotlib.axes.Axes.axis`: Axis on which network plot is drawn.

    Note:
        This function requires `matplotlib <https://matplotlib.org/>`_.
    """
    try:
        plt
    except NameError:
        raise ImportError(
            "`matplotlib` is not installed, so `textacy.viz` won't work; "
            "install it individually via `$ pip install matplotlib`, or "
            "along with textacy via `pip install textacy[viz]`."
        )
    with plt.rc_context(RC_PARAMS):
        fig, ax = plt.subplots(figsize=(12, 12))

        pos = nx.layout.spring_layout(graph, k=spread / math.sqrt(len(graph.nodes())))
        _ = nx.draw_networkx_edges(
            graph, ax=ax, pos=pos, width=line_width, alpha=line_alpha, arrows=False
        )

        if node_weights is None:
            if draw_nodes is True:
                _ = nx.draw_networkx_nodes(
                    graph,
                    ax=ax,
                    pos=pos,
                    alpha=node_alpha,
                    linewidths=0.5,
                    node_size=base_node_size,
                )
            _ = nx.draw_networkx_labels(
                graph,
                pos,
                ax=ax,
                font_size=base_font_size,
                font_color="black",
                font_family="sans-serif",
            )
        else:
            max_node_weight = max(node_weights.values())
            if draw_nodes is True:
                node_sizes = [
                    base_node_size * pow(node_weights[node] / max_node_weight, 0.75)
                    for node in graph.nodes()
                ]
                _ = nx.draw_networkx_nodes(
                    graph,
                    ax=ax,
                    pos=pos,
                    node_size=node_sizes,
                    alpha=node_alpha,
                    linewidths=0.5,
                )
            for node, weight in node_weights.items():
                _ = nx.draw_networkx_labels(
                    graph,
                    pos,
                    labels={node: node},
                    ax=ax,
                    font_color="black",
                    font_family="sans-serif",
                    font_size=base_font_size * pow(weight / max_node_weight, 0.15),
                )

        ax.set_frame_on(False)
        ax.set_xticklabels(["" for _ in range(len(ax.get_xticklabels()))])
        ax.set_yticklabels(["" for _ in range(len(ax.get_yticklabels()))])

    if save:
        fig.savefig(save, bbox_inches="tight", dpi=100)

    return ax



"""
Replace
-------

Replace parts of raw text that are semantically important as members of a group
but not so much in the individual instances.
"""
from .resources import (
    RE_CURRENCY_SYMBOL,
    RE_EMAIL,
    RE_EMOJI,
    RE_HASHTAG,
    RE_NUMBER,
    RE_PHONE_NUMBER,
    RE_SHORT_URL,
    RE_URL,
    RE_USER_HANDLE,
)


def replace_currency_symbols(text: str, replace_with: str = "_CUR_") -> str:
    """Replace all currency symbols in ``text`` with ``replace_with``."""
    return RE_CURRENCY_SYMBOL.sub(replace_with, text)


def replace_emails(text: str, replace_with: str = "_EMAIL_") -> str:
    """Replace all email addresses in ``text`` with ``replace_with``."""
    return RE_EMAIL.sub(replace_with, text)


def replace_emojis(text: str, replace_with: str = "_EMOJI_") -> str:
    """
    Replace all emoji and pictographs in ``text`` with ``replace_with``.

    Note:
        If your Python has a narrow unicode build ("USC-2"), only dingbats
        and miscellaneous symbols are replaced because Python isn't able
        to represent the unicode data for things like emoticons. Sorry!
    """
    return RE_EMOJI.sub(replace_with, text)


def replace_hashtags(text: str, replace_with: str = "_TAG_") -> str:
    """Replace all hashtags in ``text`` with ``replace_with``."""
    return RE_HASHTAG.sub(replace_with, text)


def replace_numbers(text: str, replace_with: str = "_NUMBER_") -> str:
    """Replace all numbers in ``text`` with ``replace_with``."""
    return RE_NUMBER.sub(replace_with, text)


def replace_phone_numbers(text: str, replace_with: str = "_PHONE_") -> str:
    """Replace all phone numbers in ``text`` with ``replace_with``."""
    return RE_PHONE_NUMBER.sub(replace_with, text)


def replace_urls(text: str, replace_with: str = "_URL_") -> str:
    """Replace all URLs in ``text`` with ``replace_with``."""
    return RE_SHORT_URL.sub(replace_with, RE_URL.sub(replace_with, text))


def replace_user_handles(text: str, replace_with: str = "_USER_") -> str:
    """Replace all user handles in ``text`` with ``replace_with``."""
    return RE_USER_HANDLE.sub(replace_with, text)



import inspect
import random
from typing import Callable, Optional, Sequence, Union

from spacy.tokens import Doc, Span

from .. import spacier
from . import utils as aug_utils


class Augmenter:
    """
    Randomly apply one or many data augmentation transforms to spaCy ``Doc`` s
    to produce new docs with additional variety and/or noise in the data.

    Initialize an ``Augmenter`` with multiple transforms, and customize the randomization
    of their selection when applying to a document::

        >>> tfs = [transforms.delete_words, transforms.swap_chars, transforms.delete_chars]
        >>> Augmenter(tfs, num=None)  # all tfs applied each time
        >>> Augmenter(tfs, num=1)  # one randomly-selected tf applied each time
        >>> Augmenter(tfs, num=0.5)  # tfs randomly selected with 50% prob each time
        >>> augmenter = Augmenter(tfs, num=[0.4, 0.8, 0.6])  # tfs randomly selected with 40%, 80%, 60% probs, respectively, each time

    Apply transforms to a given ``Doc`` to produce new documents::

        >>> text = "The quick brown fox jumps over the lazy dog."
        >>> doc = textacy.make_spacy_doc(text, lang="en")
        >>> augmenter.apply_transforms(doc)
        The quick brown ox jupms over the lazy dog.
        >>> augmenter.apply_transforms(doc)
        The quikc brown fox over the lazy dog.
        >>> augmenter.apply_transforms(doc)
        quick brown fox jumps over teh lazy dog.

    Parameters for individual transforms may be specified when initializing ``Augmenter``
    or, if necessary, when applying to individual documents::

        >>> from functools import partial
        >>> tfs = [partial(transforms.delete_words, num=3), transforms.swap_chars]
        >>> augmenter = Augmenter(tfs)
        >>> augmenter.apply_transforms(doc)
        brown fox jumps over layz dog.
        >>> augmenter.apply_transforms(doc, lang=doc.lang)  # (not actually needed for these tfs)
        quick brown fox over teh lazy.

    Args:
        transforms: Ordered sequence of callables that must take List[:obj:`AugTok`]
            as their first positional argument and return another List[:obj:`AugTok`].

            .. note:: Although the particular transforms applied may vary doc-by-doc,
               they are applied *in order* as listed here. Since some transforms may
               clobber text in a way that makes other transforms less effective,
               a stable ordering can improve the quality of augmented data.

        num: If int, number of transforms to randomly select from ``transforms`` each time
            :meth:`Augmenter.apply_tranforms()` is called.
            If float, probability that any given transform will be selected.
            If Sequence[float], the probability that the corresponding transform
            in ``transforms`` will be selected (these must be the same length).
            If None (default), num is set to ``len(transforms)``, which means that
            every transform is applied each time.

    See Also:
        A collection of general-purpose transforms are implemented in
        :mod:`textacy.augmentation.transforms`.
    """

    def __init__(
        self,
        transforms: Sequence[Callable],
        *,
        num: Optional[Union[int, float, Sequence[float]]] = None,
    ):
        self.tfs = self._validate_transforms(transforms)
        self.num = self._validate_num(num)

    def apply_transforms(self, doc: Doc, **kwargs) -> Doc:
        """
        Sequentially apply some subset of data augmentation transforms to ``doc``,
        then return a new ``Doc`` created from the augmented text.

        Args:
            doc
            **kwargs: If, for whatever reason, you have to pass keyword argument values
                into transforms that vary or depend on characteristics of ``doc``,
                specify them here. The transforms' call signatures will be inspected,
                and values will be passed along, as needed.

        Returns:
            :class:`spacy.tokens.Doc`
        """
        if doc.is_sentenced:
            nested_aug_toks = [aug_utils.to_aug_toks(sent) for sent in doc.sents]
        else:
            nested_aug_toks = [aug_utils.to_aug_toks(doc)]
        lang = kwargs.get("lang") or doc.vocab.lang
        tfs = self._get_random_transforms()
        new_nested_aug_toks = []
        for aug_toks in nested_aug_toks:
            # this is a bit of a hack, but whatchagonnado
            if kwargs:
                for tf in tfs:
                    tf_params = inspect.signature(tf).parameters
                    tf_kwargs = {
                        kwarg: value for kwarg, value in kwargs.items()
                        if kwarg in tf_params
                    }
                    aug_toks = tf(aug_toks, **tf_kwargs)
            else:
                for tf in tfs:
                    aug_toks = tf(aug_toks)
            new_nested_aug_toks.append(aug_toks)
        return self._make_new_spacy_doc(new_nested_aug_toks, lang)

    def _validate_transforms(self, transforms):
        transforms = tuple(transforms)
        if not transforms:
            raise ValueError("at least one transform callable must be specified")
        elif not all(callable(transform) for transform in transforms):
            raise TypeError("all transforms must be callable")
        else:
            return transforms

    def _validate_num(self, num):
        if num is None:
            return len(self.tfs)
        elif isinstance(num, int) and 0 <= num <= len(self.tfs):
            return num
        elif isinstance(num, float) and 0.0 <= num <= 1.0:
            return num
        elif (
            isinstance(num, (tuple, list)) and
            len(num) == len(self.tfs) and
            all(isinstance(n, float) and 0.0 <= n <= 1.0 for n in num)
        ):
            return tuple(num)
        else:
            raise ValueError(
                "num={} is invalid; must be an int >= 1, a float in [0.0, 1.0], "
                "or a list of floats of length equal to given transforms".format(num)
            )

    def _get_random_transforms(self):
        num = self.num
        if isinstance(num, int):
            rand_idxs = random.sample(range(len(self.tfs)), min(num, len(self.tfs)))
            rand_tfs = [self.tfs[idx] for idx in sorted(rand_idxs)]
        elif isinstance(num, float):
            rand_tfs = [tf for tf in self.tfs if random.random() < num]
        else:
            rand_tfs = [
                tf for tf, tf_num in zip(self.tfs, self.num)
                if random.random() < tf_num
            ]
        return rand_tfs

    def _make_new_spacy_doc(self, nested_aug_tokens, lang):
        # TODO: maybe collect words, spaces, and array vals
        # then directly instantiate a new Doc object?
        # this would require adding an array field to AugTok
        new_text = "".join(
            aug_tok.text + aug_tok.ws
            for aug_toks in nested_aug_tokens
            for aug_tok in aug_toks
        )
        return spacier.core.make_spacy_doc(new_text, lang=lang)



"""
Collection of regular expressions and other (small, generally useful) constants.
"""
import pathlib
import re
from typing import Dict, Pattern, Set


DEFAULT_DATA_DIR: pathlib.Path = pathlib.Path(__file__).parent.resolve() / "data"

NUMERIC_ENT_TYPES: Set[str] = {"ORDINAL", "CARDINAL", "MONEY", "QUANTITY", "PERCENT", "TIME", "DATE"}
SUBJ_DEPS: Set[str] = {"agent", "csubj", "csubjpass", "expl", "nsubj", "nsubjpass"}
OBJ_DEPS: Set[str] = {"attr", "dobj", "dative", "oprd"}
AUX_DEPS: Set[str] = {"aux", "auxpass", "neg"}

REPORTING_VERBS: Set[str] = {
    "according", "accuse", "acknowledge", "add", "admit", "agree",
    "allege", "announce", "argue", "ask", "assert", "believe", "blame",
    "charge", "cite", "claim", "complain", "concede", "conclude",
    "confirm", "contend", "criticize", "declare", "decline", "deny",
    "describe", "disagree", "disclose", "estimate", "explain", "fear",
    "hope", "insist", "maintain", "mention", "note", "observe", "order",
    "predict", "promise", "recall", "recommend", "reply", "report", "say",
    "state", "stress", "suggest", "tell", "testify", "think", "urge", "warn",
    "worry", "write"
}

MATCHER_VALID_OPS: Set[str] = {"!", "+", "?", "*"}

POS_REGEX_PATTERNS: Dict[str, Dict[str, str]] = {
    "en": {
        "NP": r"<DET>? <NUM>* (<ADJ> <PUNCT>? <CONJ>?)* (<NOUN>|<PROPN> <PART>?)+",
        "PP": r"<ADP> <DET>? <NUM>* (<ADJ> <PUNCT>? <CONJ>?)* (<NOUN> <PART>?)+",
        "VP": r"<AUX>* <ADV>* <VERB>",
    }
}

RE_MATCHER_TOKPAT_DELIM: Pattern = re.compile(r"\s+")
RE_MATCHER_SPECIAL_VAL: Pattern = re.compile(
    r"^(int|bool)\([^: ]+\)$",
    flags=re.UNICODE)

RE_ACRONYM: Pattern = re.compile(
    r"(?:^|(?<=\W))"
    r"(?:"
    r"(?:(?:(?:[A-Z]\.?)+[a-z0-9&/-]?)+(?:[A-Z][s.]?|\ds?))"
    r"|"
    r"(?:\d(?:\-?[A-Z])+)"
    r")"
    r"(?:$|(?=\W))",
    flags=re.UNICODE)

RE_LINEBREAK: Pattern = re.compile(r"(\r\n|[\n\v])+")
RE_NONBREAKING_SPACE: Pattern = re.compile(r"[^\S\n\v]+", flags=re.UNICODE)

# regexes for cleaning up crufty terms
RE_DANGLING_PARENS_TERM: Pattern = re.compile(
    r"(?:\s|^)(\()\s{1,2}(.*?)\s{1,2}(\))(?:\s|$)", flags=re.UNICODE)
RE_LEAD_TAIL_CRUFT_TERM: Pattern = re.compile(
    r"^[^\w(-]+|[^\w).!?]+$", flags=re.UNICODE)
RE_LEAD_HYPHEN_TERM: Pattern = re.compile(
    r"^-([^\W\d_])", flags=re.UNICODE)
RE_NEG_DIGIT_TERM: Pattern = re.compile(
    r"(-) (\d)", flags=re.UNICODE)
RE_WEIRD_HYPHEN_SPACE_TERM: Pattern = re.compile(
    r"(?<=[^\W\d]) (-[^\W\d])", flags=re.UNICODE)
RE_WEIRD_APOSTR_SPACE_TERM: Pattern = re.compile(
    r"([^\W\d]+) ('[a-z]{1,2}\b)", flags=re.UNICODE)



from .augmenter import Augmenter
from .transforms import (
    substitute_word_synonyms,
    insert_word_synonyms,
    swap_words,
    delete_words,
    substitute_chars,
    insert_chars,
    swap_chars,
    delete_chars,
)
from .utils import to_aug_toks, get_char_weights



"""
IMDB movie reviews
------------------

A collection of 50k highly polar movie reviews posted to IMDB, split evenly
into training and testing sets, with 25k positive and 25k negative sentiment labels,
as well as some unlabeled reviews.

Records include the following key fields (plus a few others):

    - ``text``: Full text of the review.
    - ``subset``: Subset of the dataset ("train" or "test") into which
      the review has been split.
    - ``label``: Sentiment label ("pos" or "neg") assigned to the review.
    - ``rating``: Numeric rating assigned by the original reviewer, ranging from
      1 to 10. Reviews with a rating <= 5 are "neg"; the rest are "pos".
    - ``movie_id``: Unique identifier for the movie under review within IMDB,
      useful for grouping reviews or joining with an external movie dataset.

Reference: Andrew L. Maas, Raymond E. Daly, Peter T. Pham, Dan Huang, Andrew Y. Ng,
and Christopher Potts. (2011). Learning Word Vectors for Sentiment Analysis.
The 49th Annual Meeting of the Association for Computational Linguistics (ACL 2011).
"""
import io
import itertools
import logging
import os
import re
from typing import Any, Dict, Iterable, Optional, Tuple, Union

from .. import constants, utils
from .. import io as tio
from .base import Dataset

LOGGER = logging.getLogger(__name__)

NAME = "imdb"
META = {
    "site_url": "http://ai.stanford.edu/~amaas/data/sentiment",
    "description": (
        "Collection of 50k highly polar movie reviews split evenly "
        "into train and test sets, with 25k positive and 25k negative labels. "
        "Also includes some unlabeled reviews."
    )
}
DOWNLOAD_URL = "http://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz"

RE_MOVIE_ID = re.compile(r"/(tt\d+)/")


class IMDB(Dataset):
    """
    Stream a collection of IMDB movie reviews from text files on disk,
    either as texts or text + metadata pairs.

    Download the data (one time only!), saving and extracting its contents to disk::

        >>> import textacy.datasets
        >>> ds = textacy.datasets.IMDB()
        >>> ds.download()
        >>> ds.info
        {'name': 'imdb',
         'site_url': 'http://ai.stanford.edu/~amaas/data/sentiment',
         'description': 'Collection of 50k highly polar movie reviews split evenly into train and test sets, with 25k positive and 25k negative labels. Also includes some unlabeled reviews.'}

    Iterate over movie reviews as texts or records with both text and metadata::

        >>> for text in ds.texts(limit=5):
        ...     print(text)
        >>> for text, meta in ds.records(limit=5):
        ...     print("\\n{} {}\\n{}".format(meta["label"], meta["rating"], text))

    Filter movie reviews by a variety of metadata fields and text length::

        >>> for text, meta in ds.records(label="pos", limit=5):
        ...     print(meta["rating"], ":", text)
        >>> for text, meta in ds.records(rating_range=(9, 11), limit=5):
        ...     print(meta["rating"], text)
        >>> for text in ds.texts(min_len=1000, limit=5):
        ...     print(len(text))

    Stream movie reviews into a :class:`textacy.Corpus <textacy.corpus.Corpus>`::

        >>> textacy.Corpus("en", data=ds.records(limit=100))
        Corpus(100 docs; 24340 tokens)

    Args:
        data_dir: Path to directory on disk under which the data is stored,
            i.e. ``/path/to/data_dir/imdb``.

    Attributes:
        full_rating_range: Lowest and highest ratings for which movie reviews are available.
    """

    full_rating_range: Tuple[int, int] = (1, 10)

    def __init__(
        self,
        data_dir=constants.DEFAULT_DATA_DIR.joinpath(NAME),
    ):
        super().__init__(NAME, meta=META)
        self.data_dir = utils.to_path(data_dir).resolve()
        self._movie_ids = {"train": {}, "test": {}}
        self._subset_labels = {
            "train": ("pos", "neg", "unsup"),
            "test": ("pos", "neg"),
        }
        self._subset = None
        self._label = None

    def download(self, *, force: bool = False) -> None:
        """
        Download the data as a compressed tar archive file, then save it to disk and
        extract its contents under the ``data_dir`` directory.

        Args:
            force: If True, always download the dataset even if it already exists
                on disk under ``data_dir``.
        """
        filepath = tio.download_file(
            DOWNLOAD_URL,
            filename="aclImdb.tar.gz",
            dirpath=self.data_dir,
            force=force,
        )
        if filepath:
            tio.unpack_archive(filepath, extract_dir=None)
        self._check_data()

    def _check_data(self):
        """Check that necessary data is found on disk, or raise an OSError."""
        data_dirpaths = (
            self.data_dir.joinpath("aclImdb", subset, label)
            for subset, labels in self._subset_labels.items()
            for label in labels
        )
        url_filepaths = (
            self.data_dir.joinpath("aclImdb", subset, "urls_{}.txt".format(label))
            for subset, labels in self._subset_labels.items()
            for label in labels
        )
        for dirpath in data_dirpaths:
            if not dirpath.is_dir():
                raise OSError(
                    "data directory {} not found; "
                    "has the dataset been downloaded?".format(dirpath)
                )
        for filepath in url_filepaths:
            if not filepath.is_file():
                raise OSError(
                    "data file {} not found; "
                    "has the dataset been downloaded?".format(filepath)
                )

    def __iter__(self):
        self._check_data()
        dirpaths = tuple(
            self.data_dir.joinpath("aclImdb", subset, label)
            for subset in self._subset or self._subset_labels.keys()
            for label in self._label or self._subset_labels[subset]
        )
        for dirpath in dirpaths:
            for filepath in tio.get_filepaths(dirpath, match_regex=r"^\d+_\d+\.txt$"):
                yield self._load_record(filepath)

    def _load_record(self, filepath: str) -> Dict[str, Any]:
        dirpath, filename = os.path.split(filepath)
        dirpath, label = os.path.split(dirpath)
        _, subset = os.path.split(dirpath)
        id_, rating = filename[:-4].split("_")
        with io.open(filepath, mode="rt", encoding="utf-8") as f:
            text = f.read().replace("<br />", "\n").strip()
        return {
            "text": text,
            "subset": subset,
            "label": label,
            "rating": int(rating) if label != "unsup" else None,
            "movie_id": self._get_movie_id(subset, label, int(id_))
        }

    def _get_movie_id(self, subset, label, id_):
        try:
            return self._movie_ids[subset][label][id_]
        except KeyError:
            fpath = self.data_dir.joinpath("aclImdb", subset, "urls_{}.txt".format(label))
            self._movie_ids[subset][label] = {
                id_: RE_MOVIE_ID.search(line).group(1)
                for id_, line in enumerate(tio.read_text(fpath, mode="rt", lines=True))
            }
            return self._movie_ids[subset][label][id_]

    def _get_filters(self, rating_range, min_len):
        filters = []
        if min_len is not None:
            if min_len < 1:
                raise ValueError("`min_len` must be at least 1")
            filters.append(
                lambda record: len(record.get("text", "")) >= min_len
            )
        if rating_range is not None:
            rating_range = utils.validate_and_clip_range(
                rating_range, self.full_rating_range, val_type=int)
            filters.append(
                lambda record: (
                    record.get("rating")
                    and rating_range[0] <= record["rating"] < rating_range[1]
                )
            )
        return filters

    def _filtered_iter(self, filters):
        if filters:
            for record in self:
                if all(filter_(record) for filter_ in filters):
                    yield record
        else:
            for record in self:
                yield record

    def texts(
        self,
        *,
        subset: Optional[str] = None,
        label: Optional[str] = None,
        rating_range: Optional[Tuple[Optional[int], Optional[int]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[str]:
        """
        Iterate over movie reviews in this dataset, optionally filtering by
        a variety of metadata and/or text length, and yield texts only.

        Args:
            subset ({"train", "test"}): Filter movie reviews
                by the dataset subset into which they've already been split.
            label ({"pos", "neg", "unsup"}): Filter movie reviews
                by the assigned sentiment label (or lack thereof, for "unsup").
            rating_range: Filter movie reviews by the rating assigned by the reviewer.
                Only those with ratings in the interval [low, high) are included.
                Both low and high values must be specified, but a null value
                for either is automatically replaced by the minimum or maximum
                valid values, respectively.
            min_len: Filter reviews by the length (# characters) of their text content.
            limit: Yield no more than ``limit`` reviews that match all specified filters.

        Yields:
            Text of the next movie review in dataset passing all filters.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        self._subset = utils.to_collection(subset, (str, bytes), tuple)
        self._label = utils.to_collection(label, (str, bytes), tuple)
        try:
            filters = self._get_filters(rating_range, min_len)
            for record in itertools.islice(self._filtered_iter(filters), limit):
                yield record["text"]
        finally:
            self._subset = None
            self._label = None

    def records(
        self,
        *,
        subset: Optional[str] = None,
        label: Optional[str] = None,
        rating_range: Optional[Tuple[Optional[int], Optional[int]]] = None,
        min_len: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Iterable[Tuple[str, dict]]:
        """
        Iterate over movie reviews in this dataset, optionally filtering by
        a variety of metadata and/or text length, and yield text + metadata pairs.

        Args:
            subset ({"train", "test"}): Filter movie reviews
                by the dataset subset into which they've already been split.
            label ({"pos", "neg", "unsup"}): Filter movie reviews
                by the assigned sentiment label (or lack thereof, for "unsup").
            rating_range: Filter movie reviews by the rating assigned by the reviewer.
                Only those with ratings in the interval [low, high) are included.
                Both low and high values must be specified, but a null value
                for either is automatically replaced by the minimum or maximum
                valid values, respectively.
            min_len: Filter reviews by the length (# characters) of their text content.
            limit: Yield no more than ``limit`` reviews that match all specified filters.

        Yields:
            Text of the next movie review in dataset passing all filters,
            and its corresponding metadata.

        Raises:
            ValueError: If any filtering options are invalid.
        """
        self._subset = utils.to_collection(subset, (str, bytes), tuple)
        self._label = utils.to_collection(label, (str, bytes), tuple)
        try:
            filters = self._get_filters(rating_range, min_len)
            for record in itertools.islice(self._filtered_iter(filters), limit):
                yield record.pop("text"), record
        finally:
            self._subset = None
            self._label = None



"""
IO Utils
--------

Functions to help read and write data to disk in a variety of formats.
"""
import bz2
import gzip
import io
import itertools
import logging
import lzma
import os
import pathlib
import re
import shutil
import tarfile
import urllib
import zipfile
from typing import cast, IO, Iterable, Iterator, Optional, Tuple, Union

from cytoolz import itertoolz

from .. import constants
from .. import utils
from .http import write_http_stream


LOGGER = logging.getLogger(__name__)

_ext_to_compression = {".bz2": "bz2", ".gz": "gzip", ".xz": "xz", ".zip": "zip"}


def open_sesame(
    filepath: Union[str, pathlib.Path],
    *,
    mode: str = "rt",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    compression: str = "infer",
    make_dirs: bool = False,
) -> IO:
    """
    Open file ``filepath``. Automatically handle file compression, relative
    paths and symlinks, and missing intermediate directory creation, as needed.

    ``open_sesame`` may be used as a drop-in replacement for :func:`io.open`.

    Args:
        filepath: Path on disk (absolute or relative) of the file to open.
        mode: The mode in which ``filepath`` is opened.
        encoding: Name of the encoding used to decode or encode ``filepath``.
            Only applicable in text mode.
        errors: String specifying how encoding/decoding errors are handled.
            Only applicable in text mode.
        newline: String specifying how universal newlines mode works.
            Only applicable in text mode.
        compression: Type of compression, if any, with which ``filepath``
            is read from or written to disk. If None, no compression is used;
            if 'infer', compression is inferrred from the extension on ``filepath``.
        make_dirs: If True, automatically create (sub)directories
            if not already present in order to write ``filepath``.

    Returns:
        file object

    Raises:
        TypeError: if ``filepath`` is not a string
        ValueError: if ``encoding`` is specified but ``mode`` is binary
        OSError: if ``filepath`` doesn't exist but ``mode`` is read
    """
    # check args
    if encoding and "t" not in mode:
        raise ValueError("encoding only applicable for text mode")

    # normalize filepath and make dirs, as needed
    filepath = utils.to_path(filepath).resolve()
    if make_dirs is True:
        _make_dirs(filepath, mode)
    elif mode.startswith("r") and not filepath.is_file():
        raise OSError("file '{}' does not exist".format(filepath))

    compression = _get_compression(filepath, compression)
    f = _get_file_handle(
        filepath,
        mode,
        compression=compression,
        encoding=encoding,
        errors=errors,
        newline=newline,
    )
    return f


def _get_compression(filepath, compression):
    """
    Get the compression method for ``filepath``, depending on its file extension
    and the value of ``compression``. Also validate the given values.
    """
    # user has specified "no compression"
    if compression is None:
        return None
    # user wants us to infer compression from filepath
    elif compression == "infer":
        ext = filepath.suffix
        try:
            return _ext_to_compression[ext.lower()]
        except KeyError:
            return None
    # user has specified compression; validate it
    elif compression in _ext_to_compression.values():
        return compression
    else:
        raise ValueError(
            "compression='{}' is invalid; "
            "valid values are {}.".format(
                compression, [None, "infer"] + sorted(_ext_to_compression.values())
            )
        )


def _get_file_handle(
    filepath, mode, *, compression=None, encoding=None, errors=None, newline=None,
):
    """
    Get a file handle for the given ``filepath`` and ``mode``, plus optional kwargs.
    """
    if compression:
        mode_ = mode.replace("b", "").replace("t", "")
        if compression == "gzip":
            f = gzip.GzipFile(filepath, mode=mode_)
        elif compression == "bz2":
            f = bz2.BZ2File(filepath, mode=mode_)
        elif compression == "xz":
            f = lzma.LZMAFile(filepath, mode=mode_)
        elif compression == "zip":
            zip_file = zipfile.ZipFile(filepath, mode=mode_)
            zip_names = zip_file.namelist()
            if len(zip_names) == 1:
                f = zip_file.open(zip_names[0])
            elif len(zip_names) == 0:
                raise ValueError("no files found in zip file '{}'".format(filepath))
            else:
                raise ValueError(
                    "{} files found in zip file '{}', "
                    "but only one file is allowed".format(len(zip_names), filepath)
                )
        else:
            raise ValueError(
                "compression='{}' is invalid; "
                "valid values are {}".format(
                    compression, [None, "infer"] + sorted(_ext_to_compression.values())
                )
            )

        if "t" in mode:
            f = io.TextIOWrapper(f, encoding=encoding, errors=errors, newline=newline)

    # no compression, file is opened as usual
    else:
        f = filepath.open(mode=mode, encoding=encoding, errors=errors, newline=newline)

    return f


def _make_dirs(filepath, mode):
    """
    If writing ``filepath`` to a directory that doesn't exist, all intermediate
    directories will be created as needed.
    """
    parent = filepath.parent
    if "w" in mode and parent:
        os.makedirs(parent, exist_ok=True)


def _validate_read_mode(mode):
    if "w" in mode or "a" in mode:
        raise ValueError(
            "mode='{}' is invalid; file must be opened in read mode".format(mode)
        )


def _validate_write_mode(mode):
    if "r" in mode:
        raise ValueError(
            "mode='{}' is invalid; file must be opened in write mode".format(mode)
        )


def coerce_content_type(content: Union[str, bytes], file_mode: str) -> Union[str, bytes]:
    """
    If the `content` to be written to file and the `file_mode` used to open it
    are incompatible (either bytes with text mode or unicode with bytes mode),
    try to coerce the content type so it can be written.
    """
    if "t" in file_mode:
        return utils.to_unicode(content)
    elif "b" in file_mode:
        return utils.to_bytes(content)
    return content


def split_records(
    items: Iterable,
    content_field: Union[str, int],
    itemwise: bool = False,
) -> Iterable:
    """
    Split records' content (text) from associated metadata, but keep them paired
    together.

    Args:
        items: An iterable of dicts, e.g. as read from disk by
            :func:`read_json(lines=True) <textacy.io.json.read_json>`,
            or an iterable of lists, e.g. as read from disk by
            :func:`read_csv() <textacy.io.csv.read_csv>`.
        content_field: If str, key in each dict item whose value is the item's content (text);
            if int, index of the value in each list item corresponding to the item's content (text).
        itemwise: If True, content + metadata are paired item-wise
            as an iterable of (content, metadata) 2-tuples; if False, content +
            metadata are paired by position in two parallel iterables in the form of
            a (iterable(content), iterable(metadata)) 2-tuple.

    Returns:
        Generator(Tuple[str, dict]): If ``itemwise`` is True and ``items`` is Iterable[dict];
        the first element in each tuple is the item's content,
        the second element is its metadata as a dictionary.

        Generator(Tuple[str, list]): If ``itemwise`` is True and ``items`` is Iterable[list];
        the first element in each tuple is the item's content,
        the second element is its metadata as a list.

        Tuple[Iterable[str], Iterable[dict]]: If ``itemwise`` is False and
        ``items`` is Iterable[dict];
        the first element of the tuple is an iterable of items' contents,
        the second is an iterable of their metadata dicts.

        Tuple[Iterable[str], Iterable[list]]: If ``itemwise`` is False and
        ``items`` is Iterable[list];
        the first element of the tuple is an iterable of items' contents,
        the second is an iterable of their metadata lists.
    """
    if itemwise is True:
        return ((item.pop(content_field), item) for item in items)
    else:
        return unzip(((item.pop(content_field), item) for item in items))


def unzip(seq: Iterable) -> Tuple:
    """
    Borrowed from ``toolz.sandbox.core.unzip``, but using cytoolz instead of toolz
    to avoid the additional dependency.
    """
    seq = iter(seq)
    # check how many iterators we need
    try:
        first = tuple(next(seq))
    except StopIteration:
        return tuple()
    # and create them
    niters = len(first)
    seqs = itertools.tee(itertoolz.cons(first, seq), niters)
    return tuple(itertools.starmap(itertoolz.pluck, enumerate(seqs)))


def get_filepaths(
    dirpath: Union[str, pathlib.Path],
    *,
    match_regex: Optional[str] = None,
    ignore_regex: Optional[str] = None,
    extension: Optional[str] = None,
    ignore_invisible: bool = True,
    recursive: bool = False,
) -> Iterable[str]:
    """
    Yield full paths of files on disk under directory ``dirpath``, optionally
    filtering for or against particular patterns or file extensions and
    crawling all subdirectories.

    Args:
        dirpath: Path to directory on disk where files are stored.
        match_regex: Regular expression pattern.
            Only files whose names match this pattern are included.
        ignore_regex: Regular expression pattern.
            Only files whose names *do not* match this pattern are included.
        extension: File extension, e.g. ".txt" or ".json".
            Only files whose extensions match are included.
        ignore_invisibl: If True, ignore invisible files,
            i.e. those that begin with a period.; otherwise, include them.
        recursive: If True, iterate recursively through subdirectories
            in search of files to include; otherwise, only return files located
            directly under ``dirpath``.

    Yields:
        Next file's name, including the full path on disk.

    Raises:
        OSError: if ``dirpath`` is not found on disk
    """
    dirpath = utils.to_path(dirpath).resolve()
    if not dirpath.is_dir():
        raise OSError("directory '{}' does not exist".format(dirpath))

    re_match = re.compile(match_regex) if match_regex else None
    re_ignore = re.compile(ignore_regex) if ignore_regex else None

    def is_good_file(dpath, fname):
        if ignore_invisible and fname.startswith("."):
            return False
        if re_match and not re_match.search(fname):
            return False
        if re_ignore and re_ignore.search(fname):
            return False
        if extension and os.path.splitext(fname)[-1] != extension:
            return False
        if not os.path.isfile(os.path.join(dpath, fname)):
            return False
        return True

    if recursive is True:
        for dirpath_, _, filenames in os.walk(dirpath):
            if ignore_invisible and dirpath_.startswith("."):
                continue
            for filename in filenames:
                if filename.startswith("."):
                    continue
                if is_good_file(dirpath_, filename):
                    yield os.path.join(dirpath_, filename)
    else:
        for subpath in dirpath.iterdir():
            if is_good_file(str(dirpath), subpath.name):
                yield str(subpath)


def download_file(
    url: str,
    *,
    filename: str = None,
    dirpath: Union[str, pathlib.Path] = constants.DEFAULT_DATA_DIR,
    force: bool = False,
) -> Optional[str]:
    """
    Download a file from ``url`` and save it to disk.

    Args:
        url: Web address from which to download data.
        filename: Name of the file to which downloaded data is saved.
            If None, a filename will be inferred from the ``url``.
        dirpath: Full path to the directory on disk
            under which downloaded data will be saved as ``filename``.
        force: If True, download the data even if it already exists at
            ``dirpath/filename``; otherwise, only download if the data doesn't
            already exist on disk.

    Returns:
        Full path of file saved to disk.
    """
    if not filename:
        filename = get_filename_from_url(url)
    filepath = utils.to_path(dirpath).resolve() / filename
    if filepath.is_file() and force is False:
        LOGGER.info(
            "file '%s' already exists and force=False; skipping download...",
            filepath,
        )
        return None
    else:
        write_http_stream(url, filepath, mode="wb", make_dirs=True)
    return str(filepath)


def get_filename_from_url(url: str) -> str:
    """
    Derive a filename from a URL's path.

    Args:
        url: URL from which to extract a filename.

    Returns:
        Filename in URL.
    """
    return os.path.basename(urllib.parse.urlparse(urllib.parse.unquote_plus(url)).path)


def unpack_archive(
    filepath: Union[str, pathlib.Path],
    *,
    extract_dir: Union[str, pathlib.Path] = None,
) -> Union[str, pathlib.Path]:
    """
    Extract data from a zip or tar archive file into a directory
    (or do nothing if the file isn't an archive).

    Args:
        filepath: Full path to file on disk from which archived contents will be extracted.
        extract_dir: Full path of the directory into which contents will be extracted.
            If not provided, the same directory as ``filepath`` is used.

    Returns:
        Path to directory of extracted contents.
    """
    filepath = utils.to_path(filepath).resolve()
    if not extract_dir:
        extract_dir = str(filepath.parent)
    filepath = str(filepath)
    os.makedirs(extract_dir, exist_ok=True)
    is_zipfile = zipfile.is_zipfile(filepath)
    is_tarfile = tarfile.is_tarfile(filepath)
    if not is_zipfile and not is_tarfile:
        LOGGER.debug("'%s' is not an archive", filepath)
        return extract_dir
    else:
        LOGGER.info("extracting data from archive file '%s'", filepath)
        shutil.unpack_archive(filepath, extract_dir=extract_dir, format=None)
        # we want to rename the unpacked directory to a consistent value
        # unfortunately, shutil doesn't pass this back to us
        # so, we get the root path of all the constituent members
        if is_zipfile:
            with zipfile.ZipFile(filepath, mode="r") as zf:
                members = zf.namelist()
        else:
            with tarfile.open(filepath, mode="r") as tf:
                members = tf.getnames()
        src_basename = os.path.commonpath(members)
        dest_basename = os.path.basename(filepath)
        if src_basename:
            while True:
                tmp, _ = os.path.splitext(dest_basename)
                if tmp == dest_basename:
                    break
                else:
                    dest_basename = tmp
            if src_basename != dest_basename:
                return shutil.move(
                    os.path.join(extract_dir, src_basename),
                    os.path.join(extract_dir, dest_basename),
                )
            else:
                return os.path.join(extract_dir, src_basename)
        else:
            return extract_dir



"""
DepecheMood
-----------

DepecheMood is a high-quality and high-coverage emotion lexicon for English and Italian
text, mapping individual terms to their emotional valences. These word-emotion weights
are inferred from crowd-sourced datasets of emotionally tagged news articles
(rappler.com for English, corriere.it for Italian).

English terms are assigned weights to eight emotions:

    - AFRAID
    - AMUSED
    - ANGRY
    - ANNOYED
    - DONT_CARE
    - HAPPY
    - INSPIRED
    - SAD

Italian terms are assigned weights to five emotions:

    - DIVERTITO (~amused)
    - INDIGNATO (~annoyed)
    - PREOCCUPATO (~afraid)
    - SODDISFATTO (~happy)
    - TRISTE (~sad)
"""
import collections
import csv
import io
import statistics

from spacy.parts_of_speech import ADJ, ADV, NOUN, VERB
from spacy.tokens import Doc, Span, Token

from .. import constants
from .. import io as tio
from .. import utils
from .base import Resource


NAME = "depeche_mood"
META = {
    "site_url": "http://www.depechemood.eu",
    "publication_url": "https://arxiv.org/abs/1810.03660",
    "description": "A simple tool to analyze the emotions evoked by a text.",
}
DOWNLOAD_URL = "https://github.com/marcoguerini/DepecheMood/releases/download/v2.0/DepecheMood_v2.0.zip"


class DepecheMood(Resource):
    """
    Interface to DepecheMood, an emotion lexicon for English and Italian text.

    Download the data (one time only!), and save its contents to disk::

        >>> import textacy.resources
        >>> rs = textacy.resources.DepecheMood(lang="en", word_rep="lemmapos")
        >>> rs.download()
        >>> rs.info
        {'name': 'depeche_mood',
         'site_url': 'http://www.depechemood.eu',
         'publication_url': 'https://arxiv.org/abs/1810.03660',
         'description': 'A simple tool to analyze the emotions evoked by a text.'}

    Access emotional valences for individual terms::

        >>> rs.get_emotional_valence("disease#n")
        {'AFRAID': 0.37093526222120465,
         'AMUSED': 0.06953745082761113,
         'ANGRY': 0.06979683067736414,
         'ANNOYED': 0.06465401081252636,
         'DONT_CARE': 0.07080580707440012,
         'HAPPY': 0.07537324330608403,
         'INSPIRED': 0.13394731320662606,
         'SAD': 0.14495008187418348}
        >>> rs.get_emotional_valence("heal#v")
        {'AFRAID': 0.060450319886187334,
         'AMUSED': 0.09284046387491741,
         'ANGRY': 0.06207816933776029,
         'ANNOYED': 0.10027622719958346,
         'DONT_CARE': 0.11259594401785,
         'HAPPY': 0.09946106491457314,
         'INSPIRED': 0.37794768332634626,
         'SAD': 0.09435012744278205}

    When passing multiple terms in the form of a List[str] or ``Span`` or ``Doc``,
    emotion weights are averaged over all terms for which weights are available::

        >>> rs.get_emotional_valence(["disease#n", "heal#v"])
        {'AFRAID': 0.215692791053696,
         'AMUSED': 0.08118895735126427,
         'ANGRY': 0.06593750000756221,
         'ANNOYED': 0.08246511900605491,
         'DONT_CARE': 0.09170087554612506,
         'HAPPY': 0.08741715411032858,
         'INSPIRED': 0.25594749826648616,
         'SAD': 0.11965010465848278}
        >>> text = "The acting was sweet and amazing, but the plot was dumb and terrible."
        >>> doc = textacy.make_spacy_doc(text, lang="en")
        >>> rs.get_emotional_valence(doc)
        {'AFRAID': 0.05272350876803627,
         'AMUSED': 0.13725054992595098,
         'ANGRY': 0.15787016147081184,
         'ANNOYED': 0.1398733360688608,
         'DONT_CARE': 0.14356943460620503,
         'HAPPY': 0.11923217912716871,
         'INSPIRED': 0.17880214720077342,
         'SAD': 0.07067868283219296}
        >>> rs.get_emotional_valence(doc[0:6])  # the acting was sweet and amazing
        {'AFRAID': 0.039790959333750785,
         'AMUSED': 0.1346884072825313,
         'ANGRY': 0.1373596223131593,
         'ANNOYED': 0.11391999698695347,
         'DONT_CARE': 0.1574819173485831,
         'HAPPY': 0.1552521762333925,
         'INSPIRED': 0.21232264216449326,
         'SAD': 0.049184278337136296}

    For good measure, here's how Italian w/o POS-tagged words looks::

        >>> rs = textacy.resources.DepecheMood(lang="it", word_rep="lemma")
        >>> rs.get_emotional_valence("amore")
        {'INDIGNATO': 0.11451408951814121,
         'PREOCCUPATO': 0.1323655108545536,
         'TRISTE': 0.18249663560400609,
         'DIVERTITO': 0.33558928569110086,
         'SODDISFATTO': 0.23503447833219815}

    Args:
        data_dir (str or :class:`pathlib.Path`): Path to directory on disk
            under which resource data is stored, i.e. ``/path/to/data_dir/depeche_mood``.
        lang ({"en", "it"}): Standard two-letter code for the language of terms
            for which emotional valences are to be retrieved.
        word_rep ({"token", "lemma", "lemmapos"}): Level of text processing used
            in computing terms' emotion weights. "token" => tokenization only;
            "lemma" => tokenization and lemmatization; "lemmapos" => tokenization,
            lemmatization, and part-of-speech tagging.
        min_freq (int): Minimum number of times that a given term must have appeared
            in the source dataset for it to be included in the emotion weights dict.
            This can be used to remove noisy terms at the expense of reducing coverage.
            Researchers observed peak performance at 10, but anywhere between
            1 and 20 is reasonable.
    """

    _lang_map = {"en": "english", "it": "italian"}
    _pos_map = {NOUN: "n", VERB: "v", ADJ: "a", ADV: "r"}
    _word_reps = ("token", "lemma", "lemmapos")

    def __init__(
        self,
        data_dir=constants.DEFAULT_DATA_DIR.joinpath(NAME),
        lang="en",
        word_rep="lemmapos",
        min_freq=3,
    ):
        super().__init__(NAME, meta=META)
        if lang not in self._lang_map:
            raise ValueError(
                "lang='{}' is invalid; valid options are {}".format(
                    lang, sorted(self._lang_map.keys()))
            )
        if word_rep not in self._word_reps:
            raise ValueError(
                "word_rep='{}' is invalid; valid options are {}".format(
                    word_rep, self._word_reps)
            )
        self.lang = lang
        self.word_rep = word_rep
        self.min_freq = min_freq
        self.data_dir = utils.to_path(data_dir).resolve()
        self._filepath = self.data_dir.joinpath(
            "DepecheMood++",
            "DepecheMood_{lang}_{word_rep}_full.tsv".format(
                lang=self._lang_map[lang], word_rep=word_rep),
        )
        self._weights = None

    @property
    def filepath(self):
        """
        str: Full path on disk for the DepecheMood tsv file
        corresponding to the ``lang`` and ``word_rep``.
        """
        if self._filepath.is_file():
            return str(self._filepath)
        else:
            return None

    @property
    def weights(self):
        """
        Dict[str, Dict[str, float]]: Mapping of term string (or term#POS,
        if :attr:`DepecheMood.word_rep` is "lemmapos") to the terms' normalized weights
        on a fixed set of affective dimensions (aka "emotions").
        """
        if not self._weights:
            if not self.filepath:
                raise OSError(
                    "resource file {} not found;\n"
                    "has the data been downloaded yet?".format(self._filepath)
                )
            with io.open(self.filepath, mode="rt", encoding="utf-8") as csvfile:
                csv_reader = csv.reader(csvfile, delimiter="\t")
                rows = list(csv_reader)
            cols = rows[0]
            self._weights = {
                row[0]: {col: float(val) for col, val in zip(cols[1:-1], row[1:-1])}
                for row in rows[1:]
                if int(row[-1]) >= self.min_freq
            }
        return self._weights

    def download(self, *, force=False):
        """
        Download resource data as a zip archive file, then save it to disk
        and extract its contents under the ``data_dir`` directory.

        Args:
            force (bool): If True, download the resource, even if it already
                exists on disk under ``data_dir``.
        """
        filepath = tio.download_file(
            DOWNLOAD_URL,
            filename=None,
            dirpath=self.data_dir,
            force=force,
        )
        if filepath:
            tio.unpack_archive(filepath, extract_dir=None)

    def get_emotional_valence(self, terms):
        """
        Get average emotional valence over all terms in ``terms`` for which
        emotion weights are available.

        Args:
            terms (str or Sequence[str], ``Token`` or Sequence[``Token``]):
                One or more terms over which to average emotional valences.
                Note that only nouns, adjectives, adverbs, and verbs are included.

                .. note:: If the resource was initialized with ``word_rep="lemmapos"``,
                   then string terms must have matching parts-of-speech appended to them
                   like TERM#POS. Only "n" => noun, "v" => verb, "a" => adjective, and
                   "r" => adverb are included in the data.

        Returns:
            Dict[str, float]: Mapping of emotion to average weight.
        """
        if isinstance(terms, (Token, str)):
            return self._get_term_emotional_valence(terms)
        elif isinstance(terms, (Span, Doc, collections.abc.Sequence)):
            return self._get_terms_emotional_valence(terms)
        else:
            raise TypeError(
                "`terms` must be of type {}, not {}".format(
                    {Token, Span, Doc, str, collections.abc.Sequence}, type(terms))
            )

    def _get_term_emotional_valence(self, term):
        """
        Args:
            term (str or :class:`spacy.tokens.Token`)

        Returns:
            Dict[str, float]
        """
        try:
            if isinstance(term, str):
                return self.weights[term]
            elif isinstance(term, Token):
                if self.word_rep == "lemmapos":
                    return self.weights["{}#{}".format(term.lemma_, self._pos_map[term.pos])]
                elif self.word_rep == "lemma":
                    return self.weights[term.lemma_]
                else:   # word_rep == "token"
                    return self.weights[term.text]
            else:
                raise TypeError(
                    "`term` must be of type {}, not {}".format({str, Token}, type(term))
                )
        except KeyError:
            return {}

    def _get_terms_emotional_valence(self, terms):
        """
        Args:
            term (Sequence[str] or Sequence[:class:`spacy.tokens.Token`])

        Returns:
            Dict[str, float]
        """
        all_emo_weights = collections.defaultdict(list)
        for term in terms:
            emo_weights = self._get_term_emotional_valence(term)
            for emo, weight in emo_weights.items():
                all_emo_weights[emo].append(weight)
        return {
            emo: statistics.mean(weights)
            for emo, weights in all_emo_weights.items()
        }



"""
Semantic Networks
-----------------

Represent documents as semantic networks, where nodes are individual terms or
whole sentences and edges are weighted by the strength of their co-occurrence or
similarity, respectively.
"""
import collections
import itertools
import logging
from typing import Callable, DefaultDict, Sequence, Union

import networkx as nx
from cytoolz import itertoolz
from spacy.tokens import Span, Token

from . import extract
# from . import vsm  # we're hiding this import within a function

LOGGER = logging.getLogger(__name__)


def terms_to_semantic_network(
    terms: Union[Sequence[str], Sequence[Token]],
    *,
    normalize: Union[str, bool, Callable[[Token], str]] = "lemma",
    window_width: int = 10,
    edge_weighting: str = "cooc_freq",
) -> nx.Graph:
    """
    Transform an ordered list of non-overlapping terms into a semantic network,
    where each term is represented by a node with weighted edges linking it to
    other terms that co-occur within ``window_width`` terms of itself.

    Args:
        terms
        normalize: If "lemma", lemmatize terms; if "lower", lowercase terms;
            if falsy, use the form of terms as they appear in ``terms``;
            if a callable, must accept a ``Token`` and return a str,
            e.g. :func:`textacy.spacier.utils.get_normalized_text()`.

            .. note:: This is applied to the elements of ``terms`` *only* if
               it's a list of ``Token``.

        window_width: Size of sliding window over ``terms`` that determines
            which are said to co-occur. If 2, only immediately adjacent terms
            have edges in the returned network.
        edge_weighting ({'cooc_freq', 'binary'}): If 'cooc_freq', the nodes for
            all co-occurring terms are connected by edges with weight equal to
            the number of times they co-occurred within a sliding window;
            if 'binary', all such edges have weight = 1.

    Returns:
        Networkx graph whose nodes represent individual terms, connected by edges
        based on term co-occurrence with weights determined by ``edge_weighting``.

    Note:
        - Be sure to filter out stopwords, punctuation, certain parts of speech, etc.
          from the terms list before passing it to this function
        - Multi-word terms, such as named entities and compound nouns, must be merged
          into single strings or ``Token`` s beforehand
        - If terms are already strings, be sure to have normalized them so that
          like terms are counted together; for example, by applying
          :func:`textacy.spacier.utils.get_normalized_text()`
    """
    if window_width < 2:
        raise ValueError(
            "`window_width` = {} is invalid; value must be >= 2".format(window_width)
        )
    if not terms:
        LOGGER.warning("input `terms` is empty, so output graph is also empty")
        return nx.Graph()

    # if len(terms) < window_width, cytoolz throws a StopIteration error
    # which we don't want
    if len(terms) < window_width:
        LOGGER.info(
            "`terms` has fewer items (%s) than the specified `window_width` (%s); "
            "setting window width to %s",
            len(terms),
            window_width,
            len(terms),
        )
        window_width = len(terms)

    if isinstance(terms[0], str):
        windows = itertoolz.sliding_window(window_width, terms)
    elif isinstance(terms[0], Token):
        if normalize == "lemma":
            windows = (
                (tok.lemma_ for tok in window)
                for window in itertoolz.sliding_window(window_width, terms)
            )
        elif normalize == "lower":
            windows = (
                (tok.lower_ for tok in window)
                for window in itertoolz.sliding_window(window_width, terms)
            )
        elif not normalize:
            windows = (
                (tok.text for tok in window)
                for window in itertoolz.sliding_window(window_width, terms)
            )
        else:
            windows = (
                (normalize(tok) for tok in window)
                for window in itertoolz.sliding_window(window_width, terms)
            )
    else:
        raise TypeError(
            "items in `terms` must be strings or spacy tokens, not {}".format(
                type(terms[0])
            )
        )

    graph = nx.Graph()

    if edge_weighting == "cooc_freq":
        cooc_mat: DefaultDict[str, DefaultDict[str, int]]
        cooc_mat = collections.defaultdict(lambda: collections.defaultdict(int))
        for window in windows:
            for w1, w2 in itertools.combinations(sorted(window), 2):
                cooc_mat[w1][w2] += 1
        graph.add_edges_from(
            (w1, w2, {"weight": weight})
            for w1, w2s in cooc_mat.items()
            for w2, weight in w2s.items()
        )
    elif edge_weighting == "binary":
        graph.add_edges_from(
            w1_w2 for window in windows for w1_w2 in itertools.combinations(window, 2)
        )

    return graph


def sents_to_semantic_network(
    sents: Union[Sequence[str], Sequence[Span]],
    *,
    normalize: Union[str, bool, Callable[[Token], str]] = "lemma",
    edge_weighting: str = "cosine",
) -> nx.Graph:
    """
    Transform a list of sentences into a semantic network, where each sentence is
    represented by a node with edges linking it to other sentences weighted by
    the (cosine or jaccard) similarity of their constituent words.

    Args:
        sents
        normalize: If 'lemma', lemmatize words in sents; if 'lower', lowercase words
        in sents; if false-y, use the form of words as they appear in sents; if a callable,
        must accept a :class:`spacy.tokens.Token` and return a str,
        e.g. :func:`textacy.spacier.utils.get_normalized_text()`.

            .. note:: This is applied to the elements of ``sents`` *only* if
               it's a list of ``Span`` s.

        edge_weighting ({'cosine', 'jaccard'}): Similarity metric to use for
            weighting edges between sentences. If 'cosine', use the cosine
            similarity between sentences represented as tf-idf word vectors;
            if 'jaccard', use the set intersection divided by the set union of
            all words in a given sentence pair.

    Returns:
        Networkx graph whose nodes are the integer indexes of the sentences in ``sents``,
        *not* the actual text of the sentences. Edges connect every node,
        with weights determined by ``edge_weighting``.

    Note:
        - If passing sentences as strings, be sure to filter out stopwords, punctuation,
          certain parts of speech, etc. beforehand
        - Consider normalizing the strings so that like terms are counted together
          (see :func:`textacy.spacier.utils.get_normalized_text()`)
    """
    from . import vsm

    if isinstance(sents[0], str):
        pass
    elif isinstance(sents[0], Span):
        if normalize == "lemma":
            sents = (
                (
                    tok.lemma_
                    for tok in extract.words(
                        sent, filter_stops=True, filter_punct=True, filter_nums=False
                    )
                )
                for sent in sents
            )
        elif normalize == "lower":
            sents = (
                (
                    tok.lower_
                    for tok in extract.words(
                        sent, filter_stops=True, filter_punct=True, filter_nums=False
                    )
                )
                for sent in sents
            )
        elif not normalize:
            sents = (
                (
                    tok.text
                    for tok in extract.words(
                        sent, filter_stops=True, filter_punct=True, filter_nums=False
                    )
                )
                for sent in sents
            )
        else:
            sents = (
                (
                    normalize(tok)
                    for tok in extract.words(
                        sent, filter_stops=True, filter_punct=True, filter_nums=False
                    )
                )
                for sent in sents
            )
    else:
        raise TypeError(
            "items in `sents` must be strings or spacy tokens, not {}".format(
                type(sents[0])
            )
        )

    if edge_weighting == "cosine":
        term_sent_matrix = vsm.Vectorizer(
            tf_type="linear", apply_idf=True, idf_type="smooth"
        ).fit_transform(sents)
    elif edge_weighting == "jaccard":
        term_sent_matrix = vsm.Vectorizer(
            tf_type="binary", apply_idf=False
        ).fit_transform(sents)
    weights = (term_sent_matrix * term_sent_matrix.T).A.tolist()
    n_sents = len(weights)

    graph = nx.Graph()
    graph.add_edges_from(
        (i, j, {"weight": weights[i][j]})
        for i in range(n_sents)
        for j in range(i + 1, n_sents)
    )

    return graph



"""
Topic Models
------------

Convenient and consolidated topic-modeling, built on ``scikit-learn``.
"""
import logging

import joblib
import numpy as np
from sklearn.decomposition import NMF, LatentDirichletAllocation, TruncatedSVD

from .. import viz

LOGGER = logging.getLogger(__name__)


class TopicModel:
    """
    Train and apply a topic model to vectorized texts using scikit-learn's
    implementations of LSA, LDA, and NMF models. Also any other topic model implementations that have
    `component_`, `n_topics` and `transform` attributes. Inspect and visualize results.
    Save and load trained models to and from disk.

    Prepare a vectorized corpus (i.e. document-term matrix) and corresponding
    vocabulary (i.e. mapping of term strings to column indices in the matrix).
    See :class:`textacy.vsm.Vectorizer` for details. In short:

    .. code-block:: pycon

        >>> vectorizer = Vectorizer(
        ...     tf_type="linear", apply_idf=True, idf_type="smooth", norm="l2",
        ...     min_df=3, max_df=0.95, max_n_terms=100000)
        >>> doc_term_matrix = vectorizer.fit_transform(terms_list)

    Initialize and train a topic model:

    .. code-block:: pycon

        >>> model = textacy.tm.TopicModel("nmf", n_topics=20)
        >>> model.fit(doc_term_matrix)
        >>> model
        TopicModel(n_topics=10, model=NMF)

    Transform the corpus and interpret our model:

    .. code-block:: pycon

        >>> doc_topic_matrix = model.transform(doc_term_matrix)
        >>> for topic_idx, top_terms in model.top_topic_terms(vectorizer.id_to_term, topics=[0,1]):
        ...     print("topic", topic_idx, ":", "   ".join(top_terms))
        topic 0 : people   american   go   year   work   think   $   today   money   america
        topic 1 : rescind   quorum   order   unanimous   consent   ask   president   mr.   madam   absence
        >>> for topic_idx, top_docs in model.top_topic_docs(doc_topic_matrix, topics=[0,1], top_n=2):
        ...     print(topic_idx)
        ...     for j in top_docs:
        ...         print(corpus[j]._.meta["title"])
        0
        THE MOST IMPORTANT ISSUES FACING THE AMERICAN PEOPLE
        55TH ANNIVERSARY OF THE BATTLE OF CRETE
        1
        CHEMICAL WEAPONS CONVENTION
        MFN STATUS FOR CHINA
        >>> for doc_idx, topics in model.top_doc_topics(doc_topic_matrix, docs=range(5), top_n=2):
        ...     print(corpus[doc_idx]._.meta["title"], ":", topics)
        JOIN THE SENATE AND PASS A CONTINUING RESOLUTION : (9, 0)
        MEETING THE CHALLENGE : (2, 0)
        DISPOSING OF SENATE AMENDMENT TO H.R. 1643, EXTENSION OF MOST-FAVORED- NATION TREATMENT FOR BULGARIA : (0, 9)
        EXAMINING THE SPEAKER'S UPCOMING TRAVEL SCHEDULE : (0, 9)
        FLOODING IN PENNSYLVANIA : (0, 9)
        >>> for i, val in enumerate(model.topic_weights(doc_topic_matrix)):
        ...     print(i, val)
        0 0.302796022302
        1 0.0635617650602
        2 0.0744927472417
        3 0.0905778808867
        4 0.0521162262192
        5 0.0656303769725
        6 0.0973516532757
        7 0.112907245542
        8 0.0680659204364
        9 0.0725001620636

    Visualize the model:

    .. code-block:: pycon

        >>> model.termite_plot(doc_term_matrix, vectorizer.id_to_term,
        ...                    topics=-1,  n_terms=25, sort_terms_by="seriation")

    Persist our topic model to disk:

    .. code-block:: pycon

        >>> model.save("nmf-10topics.pkl")

    Args:
        model ({"nmf", "lda", "lsa"} or ``sklearn.decomposition.<model>``)
        n_topics (int): number of topics in the model to be initialized
        **kwargs:
            variety of parameters used to initialize the model; see individual
            sklearn pages for full details

    Raises:
        ValueError: if ``model`` not in ``{"nmf", "lda", "lsa"}`` or is not an
            NMF, LatentDirichletAllocation, or TruncatedSVD instance

    See Also:
        * http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.NMF.html
        * http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.LatentDirichletAllocation.html
        * http://scikit-learn.org/stable/modules/generated/sklearn.decomposition.TruncatedSVD.html
    """

    _required_trained_model_attr = {"transform", "components_", "n_topics"}

    def __init__(self, model, n_topics=10, **kwargs):
        if isinstance(model, (NMF, LatentDirichletAllocation, TruncatedSVD)):
            self.model = model
        elif all(hasattr(model, required_attr) for required_attr in self._required_trained_model_attr):
            self.model = model
        else:
            self.init_model(model, n_topics=n_topics, **kwargs)

    def init_model(self, model, n_topics=10, **kwargs):
        if model == "nmf":
            self.model = NMF(
                n_components=n_topics,
                alpha=kwargs.get("alpha", 0.1),
                l1_ratio=kwargs.get("l1_ratio", 0.5),
                max_iter=kwargs.get("max_iter", 200),
                random_state=kwargs.get("random_state", 1),
                shuffle=kwargs.get("shuffle", False),
            )
        elif model == "lda":
            self.model = LatentDirichletAllocation(
                n_components=n_topics,
                max_iter=kwargs.get("max_iter", 10),
                random_state=kwargs.get("random_state", 1),
                learning_method=kwargs.get("learning_method", "online"),
                learning_offset=kwargs.get("learning_offset", 10.0),
                batch_size=kwargs.get("batch_size", 128),
                n_jobs=kwargs.get("n_jobs", 1),
            )
        elif model == "lsa":
            self.model = TruncatedSVD(
                n_components=n_topics,
                algorithm=kwargs.get("algorithm", "randomized"),
                n_iter=kwargs.get("n_iter", 5),
                random_state=kwargs.get("random_state", 1),
            )
        else:
            msg = 'model "{}" invalid; must be {}'.format(model, {"nmf", "lda", "lsa"})
            raise ValueError(msg)

    def __repr__(self):
        return "TopicModel(n_topics={}, model={})".format(
            self.n_topics, str(self.model).split("(", 1)[0]
        )

    def save(self, filepath):
        _ = joblib.dump(self.model, filepath, compress=3)
        LOGGER.info("%s model saved to %s", self.model, filepath)

    @classmethod
    def load(cls, filepath):
        model = joblib.load(filepath)
        n_topics = model.n_components if hasattr(model, "n_components") else model.n_topics
        return cls(model, n_topics=n_topics)

    def fit(self, doc_term_matrix):
        self.model.fit(doc_term_matrix)

    def partial_fit(self, doc_term_matrix):
        if isinstance(self.model, LatentDirichletAllocation):
            self.model.partial_fit(doc_term_matrix)
        else:
            raise TypeError("only LatentDirichletAllocation models have partial_fit")

    def transform(self, doc_term_matrix):
        return self.model.transform(doc_term_matrix)

    @property
    def n_topics(self):
        try:
            return self.model.n_components
        except AttributeError:
            return self.model.n_topics

    def get_doc_topic_matrix(self, doc_term_matrix, *, normalize=True):
        """
        Transform a document-term matrix into a document-topic matrix, where rows
        correspond to documents and columns to the topics in the topic model.

        Args:
            doc_term_matrix (array-like or sparse matrix): Corpus represented as a
                document-term matrix with shape (n_docs, n_terms). LDA expects
                tf-weighting, while NMF and LSA may do better with tfidf-weighting.
            normalize (bool): if True, the values in each row are normalized,
                i.e. topic weights on each document sum to 1

        Returns:
            :class:`numpy.ndarray`: Document-topic matrix with shape (n_docs, n_topics).
        """
        doc_topic_matrix = self.transform(doc_term_matrix)
        if normalize is True:
            return doc_topic_matrix / np.sum(doc_topic_matrix, axis=1, keepdims=True)
        else:
            return doc_topic_matrix

    def top_topic_terms(self, id2term, *, topics=-1, top_n=10, weights=False):
        """
        Get the top ``top_n`` terms by weight per topic in ``model``.

        Args:
            id2term (list(str) or dict): object that returns the term string corresponding
                to term id ``i`` through ``id2term[i]``; could be a list of strings
                where the index represents the term id, such as that returned by
                ``sklearn.feature_extraction.text.CountVectorizer.get_feature_names()``,
                or a mapping of term id: term string
            topics (int or Sequence[int]): topic(s) for which to return top terms;
                if -1 (default), all topics' terms are returned
            top_n (int): number of top terms to return per topic
            weights (bool): if True, terms are returned with their corresponding
                topic weights; otherwise, terms are returned without weights

        Yields:
            Tuple[int, Tuple[str]] or Tuple[int, Tuple[Tuple[str, float]]]:
                next tuple corresponding to a topic; the first element is the topic's
                index; if ``weights`` is False, the second element is a tuple of str
                representing the top ``top_n`` related terms; otherwise, the second
                is a tuple of (str, float) pairs representing the top ``top_n``
                related terms and their associated weights wrt the topic; for example::

                    >>> list(TopicModel.top_topic_terms(id2term, topics=(0, 1), top_n=2, weights=False))
                    [(0, ('foo', 'bar')), (1, ('bat', 'baz'))]
                    >>> list(TopicModel.top_topic_terms(id2term, topics=0, top_n=2, weights=True))
                    [(0, (('foo', 0.1415), ('bar', 0.0986)))]
        """
        if topics == -1:
            topics = range(self.n_topics)
        elif isinstance(topics, int):
            topics = (topics,)

        for topic_idx in topics:
            topic = self.model.components_[topic_idx]
            if weights is False:
                yield (
                    topic_idx,
                    tuple(id2term[i] for i in np.argsort(topic)[: -top_n - 1 : -1]),
                )
            else:
                yield (
                    topic_idx,
                    tuple(
                        (id2term[i], topic[i])
                        for i in np.argsort(topic)[: -top_n - 1 : -1]
                    ),
                )

    def top_topic_docs(self, doc_topic_matrix, *, topics=-1, top_n=10, weights=False):
        """
        Get the top ``top_n`` docs by weight per topic in ``doc_topic_matrix``.

        Args:
            doc_topic_matrix (:class:`numpy.ndarray`): document-topic matrix with shape
                (n_docs, n_topics), the result of calling :meth:`TopicModel.get_doc_topic_matrix()`
            topics (int or Sequence[int]): topic(s) for which to return top docs;
                if -1, all topics' docs are returned
            top_n (int): number of top docs to return per topic
            weights (bool): if True, docs are returned with their corresponding
                (normalized) topic weights; otherwise, docs are returned without weights

        Yields:
            Tuple[int, Tuple[int]] or Tuple[int, Tuple[Tuple[int, float]]]:
                next tuple corresponding to a topic; the first element is the topic's
                index; if ``weights`` is False, the second element is a tuple of ints
                representing the top ``top_n`` related docs; otherwise, the second
                is a tuple of (int, float) pairs representing the top ``top_n``
                related docs and their associated weights wrt the topic; for example::

                    >>> list(TopicModel.top_doc_terms(dtm, topics=(0, 1), top_n=2, weights=False))
                    [(0, (4, 2)), (1, (1, 3))]
                    >>> list(TopicModel.top_doc_terms(dtm, topics=0, top_n=2, weights=True))
                    [(0, ((4, 0.3217), (2, 0.2154)))]
        """
        if topics == -1:
            topics = range(self.n_topics)
        elif isinstance(topics, int):
            topics = (topics,)

        for topic_idx in topics:
            top_doc_idxs = np.argsort(doc_topic_matrix[:, topic_idx])[: -top_n - 1 : -1]
            if weights is False:
                yield (topic_idx, tuple(doc_idx for doc_idx in top_doc_idxs))
            else:
                yield (
                    topic_idx,
                    tuple(
                        (doc_idx, doc_topic_matrix[doc_idx, topic_idx])
                        for doc_idx in top_doc_idxs
                    ),
                )

    def top_doc_topics(self, doc_topic_matrix, *, docs=-1, top_n=3, weights=False):
        """
        Get the top ``top_n`` topics by weight per doc for ``docs`` in ``doc_topic_matrix``.

        Args:
            doc_topic_matrix (:class:`numpy.ndarray`): document-topic matrix with shape
                (n_docs, n_topics), the result of calling :meth:`TopicModel.get_doc_topic_matrix()`
            docs (int or Sequence[int]): docs for which to return top topics;
                if -1, all docs' top topics are returned
            top_n (int): number of top topics to return per doc
            weights (bool): if True, docs are returned with their corresponding
                (normalized) topic weights; otherwise, docs are returned without weights

        Yields:
            Tuple[int, Tuple[int]] or Tuple[int, Tuple[Tuple[int, float]]]:
                next tuple corresponding to a doc; the first element is the doc's
                index; if ``weights`` is False, the second element is a tuple of ints
                representing the top ``top_n`` related topics; otherwise, the second
                is a tuple of (int, float) pairs representing the top ``top_n``
                related topics and their associated weights wrt the doc; for example::

                    >>> list(TopicModel.top_doc_topics(dtm, docs=(0, 1), top_n=2, weights=False))
                    [(0, (1, 4)), (1, (3, 2))]
                    >>> list(TopicModel.top_doc_topics(dtm, docs=0, top_n=2, weights=True))
                    [(0, ((1, 0.2855), (4, 0.2412)))]
        """
        if docs == -1:
            docs = range(doc_topic_matrix.shape[0])
        elif isinstance(docs, int):
            docs = (docs,)

        for doc_idx in docs:
            top_topic_idxs = np.argsort(doc_topic_matrix[doc_idx, :])[: -top_n - 1 : -1]
            if weights is False:
                yield (doc_idx, tuple(topic_idx for topic_idx in top_topic_idxs))
            else:
                yield (
                    doc_idx,
                    tuple(
                        (topic_idx, doc_topic_matrix[doc_idx, topic_idx])
                        for topic_idx in top_topic_idxs
                    ),
                )

    def topic_weights(self, doc_topic_matrix):
        """
        Get the overall weight of topics across an entire corpus. Note: Values depend
        on whether topic weights per document in ``doc_topic_matrix`` were normalized,
        or not. I suppose either way makes sense... o_O

        Args:
            doc_topic_matrix (:class:`numpy.ndarray`): document-topic matrix with shape
                (n_docs, n_topics), the result of calling :meth:`TopicModel.get_doc_topic_matrix()`

        Returns:
            :class:`numpy.ndarray`: the ith element is the ith topic's overall weight
        """
        return doc_topic_matrix.sum(axis=0) / doc_topic_matrix.sum(axis=0).sum()

    # def get_topic_coherence(self, topic_idx):
    #     raise NotImplementedError()
    #
    # def get_model_coherence(self):
    #     raise NotImplementedError()

    def termite_plot(
        self,
        doc_term_matrix,
        id2term,
        *,
        topics=-1,
        sort_topics_by="index",
        highlight_topics=None,
        n_terms=25,
        rank_terms_by="topic_weight",
        sort_terms_by="seriation",
        save=False,
        rc_params=None,
    ):
        """
        Make a "termite" plot for assessing topic models using a tabular layout
        to promote comparison of terms both within and across topics.

        Args:
            doc_term_matrix (:class:`numpy.ndarray` or sparse matrix): corpus
                represented as a document-term matrix with shape (n_docs, n_terms);
                may have tf- or tfidf-weighting
            id2term (List[str] or dict): object that returns the term string corresponding
                to term id ``i`` through ``id2term[i]``; could be a list of strings
                where the index represents the term id, such as that returned by
                ``sklearn.feature_extraction.text.CountVectorizer.get_feature_names()``,
                or a mapping of term id: term string
            topics (int or Sequence[int]): topic(s) to include in termite plot;
                if -1, all topics are included
            sort_topics_by ({'index', 'weight'}):
            highlight_topics (int or Sequence[int]): indices for up to 6 topics
                to visually highlight in the plot with contrasting colors
            n_terms (int): number of top terms to include in termite plot
            rank_terms_by ({'topic_weight', 'corpus_weight'}): value used
                to rank terms; the top-ranked ``n_terms`` are included in the plot
            sort_terms_by ({'seriation', 'weight', 'index', 'alphabetical'}):
                method used to vertically sort the selected top ``n_terms`` terms;
                the default ("seriation") groups similar terms together, which
                facilitates cross-topic assessment
            save (str): give the full /path/to/fname on disk to save figure
                rc_params (dict, optional): allow passing parameters to rc_context in matplotlib.plyplot,
                details in https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.rc_context.html

        Returns:
            ``matplotlib.axes.Axes.axis``: Axis on which termite plot is plotted.

        Raises:
            ValueError: if more than 6 topics are selected for highlighting, or
                an invalid value is passed for the sort_topics_by, rank_terms_by,
                and/or sort_terms_by params

        References:
            - Chuang, Jason, Christopher D. Manning, and Jeffrey Heer. "Termite:
              Visualization techniques for assessing textual topic models."
              Proceedings of the International Working Conference on Advanced
              Visual Interfaces. ACM, 2012.
            - for sorting by "seriation", see https://arxiv.org/abs/1406.5370

        See Also:
            :func:`viz.termite_plot <textacy.viz.termite.termite_plot>`

        TODO: `rank_terms_by` other metrics, e.g. topic salience or relevance
        """
        if highlight_topics is not None:
            if isinstance(highlight_topics, int):
                highlight_topics = (highlight_topics,)
            elif len(highlight_topics) > 6:
                raise ValueError("no more than 6 topics may be highlighted at once")

        # get topics indices
        if topics == -1:
            topic_inds = tuple(range(self.n_topics))
        elif isinstance(topics, int):
            topic_inds = (topics,)
        else:
            topic_inds = tuple(topics)

        # get topic indices in sorted order
        if sort_topics_by == "index":
            topic_inds = sorted(topic_inds)
        elif sort_topics_by == "weight":
            topic_inds = tuple(
                topic_ind
                for topic_ind in np.argsort(
                    self.topic_weights(self.transform(doc_term_matrix))
                )[::-1]
                if topic_ind in topic_inds
            )
        else:
            msg = "invalid sort_topics_by value; must be in {}".format(
                {"index", "weight"}
            )
            raise ValueError(msg)

        # get column index of any topics to highlight in termite plot
        if highlight_topics is not None:
            highlight_cols = tuple(
                i
                for i in range(len(topic_inds))
                if topic_inds[i] in highlight_topics
            )
        else:
            highlight_cols = None

        # get top term indices
        if rank_terms_by == "corpus_weight":
            term_inds = np.argsort(np.ravel(doc_term_matrix.sum(axis=0)))[
                : -n_terms - 1 : -1
            ]
        elif rank_terms_by == "topic_weight":
            term_inds = np.argsort(self.model.components_.sum(axis=0))[
                : -n_terms - 1 : -1
            ]
        else:
            msg = "invalid rank_terms_by value; must be in {}".format(
                {"corpus_weight", "topic_weight"}
            )
            raise ValueError(msg)

        # get top term indices in sorted order
        if sort_terms_by == "weight":
            pass
        elif sort_terms_by == "index":
            term_inds = sorted(term_inds)
        elif sort_terms_by == "alphabetical":
            term_inds = sorted(term_inds, key=lambda x: id2term[x])
        elif sort_terms_by == "seriation":
            topic_term_weights_mat = np.array(
                np.array(
                    [
                        self.model.components_[topic_ind][term_inds]
                        for topic_ind in topic_inds
                    ]
                )
            ).T
            # calculate similarity matrix
            topic_term_weights_sim = np.dot(
                topic_term_weights_mat, topic_term_weights_mat.T
            )
            # substract minimum of sim mat in order to keep sim mat nonnegative
            topic_term_weights_sim = (
                topic_term_weights_sim - topic_term_weights_sim.min()
            )
            # compute Laplacian matrice and its 2nd eigenvector
            L = np.diag(sum(topic_term_weights_sim, 1)) - topic_term_weights_sim
            D, V = np.linalg.eigh(L)
            D = D[np.argsort(D)]
            V = V[:, np.argsort(D)]
            fiedler = V[:, 1]
            # get permutation corresponding to sorting the 2nd eigenvector
            term_inds = [term_inds[i] for i in np.argsort(fiedler)]
        else:
            msg = "invalid sort_terms_by value; must be in {}".format(
                {"weight", "index", "alphabetical", "seriation"}
            )
            raise ValueError(msg)

        # get topic and term labels
        topic_labels = tuple("topic {}".format(topic_ind) for topic_ind in topic_inds)
        term_labels = tuple(id2term[term_ind] for term_ind in term_inds)

        # get topic-term weights to size dots
        term_topic_weights = np.array(
            [self.model.components_[topic_ind][term_inds] for topic_ind in topic_inds]
        ).T

        return viz.draw_termite_plot(
            term_topic_weights,
            topic_labels,
            term_labels,
            highlight_cols=highlight_cols,
            save=save,
            rc_params=rc_params,
        )



from .utils import (
    coerce_content_type,
    download_file,
    get_filename_from_url,
    get_filepaths,
    open_sesame,
    split_records,
    unpack_archive,
    unzip,
)
from .csv import read_csv, write_csv
from .http import read_http_stream, write_http_stream
from .json import read_json, read_json_mash, write_json
from .matrix import read_sparse_matrix, write_sparse_matrix
from .spacy import read_spacy_docs, write_spacy_docs
from .text import read_text, write_text



"""
Text
----

Functions for reading from and writing to disk records in plain text format,
either as one text per file or one text per *line* in a file.
"""
import pathlib
from typing import Iterable, Optional, Union

from . import utils as io_utils


def read_text(
    filepath: Union[str, pathlib.Path],
    *,
    mode: str = "rt",
    encoding: Optional[str] = None,
    lines: bool = False,
) -> Iterable[str]:
    """
    Read the contents of a text file at ``filepath``, either all at once
    or streaming line-by-line.

    Args:
        filepath: Path to file on disk from which data will be read.
        mode: Mode with which ``filepath`` is opened.
        encoding: Name of the encoding used to decode or encode the data in ``filepath``.
            Only applicable in text mode.
        lines: If False, all data is read in at once;
            otherwise, data is read in one line at a time.

    Yields:
        Next line of text to read in.

        If ``lines`` is False, wrap this output in :func:`next()` to conveniently
        access the full text.
    """
    io_utils._validate_read_mode(mode)
    with io_utils.open_sesame(filepath, mode=mode, encoding=encoding) as f:
        if lines is False:
            yield f.read()
        else:
            for line in f:
                yield line


def write_text(
    data: Union[str, Iterable[str]],
    filepath: Union[str, pathlib.Path],
    *,
    mode: str = "wt",
    encoding: Optional[str] = None,
    make_dirs: bool = False,
    lines: bool = False,
) -> None:
    """
    Write text ``data`` to disk at ``filepath``, either all at once
    or streaming line-by-line.

    Args:
        data If ``lines`` is False, a single string to write to disk; for example::

                "isnt rick and morty that thing you get when you die and your body gets all stiff"

            If ``lines`` is True, an iterable of strings to write to disk, one
            item per line; for example::

                ["isnt rick and morty that thing you get when you die and your body gets all stiff",
                 "You're thinking of rigor mortis. Rick and morty is when you get trolled into watching \"never gonna give you up\"",
                 "That's rickrolling. Rick and morty is a type of pasta"]

        filepath: Path to file on disk to which data will be written.
        mode: Mode with which ``filepath`` is opened.
        encoding: Name of the encoding used to decode or encode the data in ``filepath``.
            Only applicable in text mode.
        make_dirs: If True, automatically create (sub)directories
            if not already present in order to write ``filepath``.
        lines: If False, all data is written at once;
            otherwise, data is written to disk one line at a time.
    """
    io_utils._validate_write_mode(mode)
    with io_utils.open_sesame(filepath, mode=mode, encoding=encoding, make_dirs=make_dirs) as f:
        if lines is False:
            f.write(data)
        else:
            newline = "\n" if "t" in mode else b"\n"
            for line in data:
                f.write(line + newline)



"""
Language identification
-----------------------

Pipeline for identifying the language of a text, using a model inspired by
Google's Compact Language Detector v3 (https://github.com/google/cld3) and
implemented with ``scikit-learn>=0.20``.

Model
^^^^^

Character unigrams, bigrams, and trigrams are extracted from input text, and
their frequencies of occurence within the text are counted. The full set of ngrams
are then hashed into a 4096-dimensional feature vector with values given by
the L2 norm of the counts. These features are passed into a Multi-layer Perceptron
with a single hidden layer of 512 rectified linear units and a softmax output layer
giving probabilities for ~140 different languages as ISO 639-1 language codes.

Technically, the model was implemented as a :class:`sklearn.pipeline.Pipeline`
with two steps: a :class:`sklearn.feature_extraction.text.HashingVectorizer`
for vectorizing input texts and a :class:`sklearn.neural_network.MLPClassifier`
for multi-class language classification.

Dataset
^^^^^^^

The pipeline was trained on a randomized, stratified subset of ~750k texts
drawn from several sources:

- **Tatoeba:** A crowd-sourced collection of sentences and their translations into
  many languages. Style is relatively informal; subject matter is a variety of
  everyday things and goings-on.
  Source: https://tatoeba.org/eng/downloads.
- **Leipzig Corpora:** A collection of corpora for many languages pulling from
  comparable sources -- specifically, 10k Wikipedia articles from official database dumps
  and 10k news articles from either RSS feeds or web scrapes, when available.
  Style is relatively formal; subject matter is a variety of notable things and goings-on.
  Source: http://wortschatz.uni-leipzig.de/en/download
- **UDHR:** The UN's Universal Declaration of Human Rights document, translated into
  hundreds of languages and split into paragraphs. Style is formal; subject matter is
  fundamental human rights to be universally protected.
  Source: https://unicode.org/udhr/index.html
- **Twitter:** A collection of tweets in each of ~70 languages, posted in July 2014,
  with languages assigned through a combination of models and human annotators.
  Style is informal; subject matter is whatever Twitter was going on about back then.
  Source: https://blog.twitter.com/engineering/en_us/a/2015/evaluating-language-identification-performance.html
- **DSLCC**: Two collections of short excerpts of journalistic texts in a handful
  of language groups that are highly similar to each other. Style is relatively formal;
  subject matter is current events.
  Source: http://ttg.uni-saarland.de/resources/DSLCC/

Performance
^^^^^^^^^^^

The trained model achieved F1 = 0.96 when (macro and micro) averaged over all languages.
A few languages have worse performance; for example, the two Norwegians ("nb" and "no"),
Bosnian ("bs") and Serbian ("sr"), and Bashkir ("ba") and Tatar ("tt") are
often confused with each other. See the textacy-data releases for more details:
https://github.com/bdewilde/textacy-data/releases/tag/lang_identifier_v1.1_sklearn_v21
"""
import io
import logging
import operator
import os
import urllib.parse

import joblib

from . import cache, constants, utils
from . import io as tio

LOGGER = logging.getLogger(__name__)


class LangIdentifier:
    """
    Args:
        data_dir (str)
        max_text_len (int)

    Attributes:
        pipeline (:class:`sklearn.pipeline.Pipeline`)
    """

    def __init__(
        self,
        data_dir=constants.DEFAULT_DATA_DIR.joinpath("lang_identifier"),
        max_text_len=1000
    ):
        self._version = 1.1
        self._model_id = self._get_model_id()
        self.data_dir = utils.to_path(data_dir).resolve()
        self.filename = self._model_id + ".pkl.gz"
        self.max_text_len = max_text_len
        self._pipeline = None

    @property
    def pipeline(self):
        if not self._pipeline:
            self._pipeline = self._load_pipeline()
        return self._pipeline

    def _load_pipeline(self):
        filepath = self.data_dir.joinpath(self.filename)
        if not filepath.is_file():
            self.download()
        with filepath.open(mode="rb") as f:
            pipeline = joblib.load(f)
        return pipeline

    def _get_model_id(self):
        fstub = "lang-identifier-v{}-sklearn-v{}"
        try:
            import pkg_resources
            filename = fstub.format(
                self._version,
                pkg_resources.get_distribution("scikit-learn").version[:4]
            )
        except ImportError:
            import sklearn
            filename = fstub.format(self._version, sklearn.__version__[:4])
        return filename

    def download(self, force=False):
        """
        Download the pipeline data as a Python version-specific compressed pickle file
        and save it to disk under the :attr:`LangIdentifier.data_dir` directory.

        Args:
            force (bool): If True, download the dataset, even if it already
                exists on disk under ``data_dir``.
        """
        release_tag = self._model_id.replace("-", "_")
        url = urllib.parse.urljoin(
            "https://github.com/bdewilde/textacy-data/releases/download/",
            release_tag + "/" + self.filename)
        tio.utils.download_file(
            url,
            filename=self.filename,
            dirpath=self.data_dir,
            force=force,
        )

    def identify_lang(self, text):
        """
        Identify the most probable language identified in ``text``.

        Args:
            text (str)

        Returns:
            str: 2-letter language code of the most probable language.
        """
        text_ = utils.to_collection(text[:self.max_text_len], str, list)
        if self._is_valid(text_[0]):
            lang = self.pipeline.predict(text_).item()
            return lang
        else:
            return "un"

    def identify_topn_langs(self, text, topn=3):
        """
        Identify the ``topn`` most probable languages identified in ``text``.

        Args:
            text (str)
            topn (int)

        Returns:
            List[Tuple[str, float]]: 2-letter language code and its probability
            for the ``topn`` most probable languages.
        """
        text_ = utils.to_collection(text[:self.max_text_len], str, list)
        if self._is_valid(text_[0]):
            lang_probs = sorted(
                zip(self.pipeline.classes_, self.pipeline.predict_proba(text_).flat),
                key=operator.itemgetter(1),
                reverse=True,
            )[:topn]
            return [(lang.item(), prob.item()) for lang, prob in lang_probs]
        else:
            return [("un", 1.0)]

    def _is_valid(self, text):
        return any(char.isalpha() for char in text)

    def init_pipeline(self):
        """
        Initialize a *new* language identification pipeline, overwriting any
        pre-trained pipeline loaded from disk under :attr:`LangIdentifier.data_dir`.
        Must be trained on (text, lang) examples before use.
        """
        import sklearn.feature_extraction
        import sklearn.neural_network
        import sklearn.pipeline

        self._pipeline = sklearn.pipeline.Pipeline(
            [
                (
                    "vectorizer",
                    sklearn.feature_extraction.text.HashingVectorizer(
                        analyzer="char_wb", ngram_range=(1, 3), lowercase=True,
                        n_features=4096, norm="l2",
                    )
                ),
                (
                    "classifier",
                    sklearn.neural_network.MLPClassifier(
                        activation="relu", solver="adam",
                        hidden_layer_sizes=(512,), alpha=0.0001, batch_size=512,
                        learning_rate_init=0.001, learning_rate="constant",
                        max_iter=15, early_stopping=True, tol=0.001,
                        shuffle=True, random_state=42,
                        verbose=True,
                    )
                ),
            ]
        )


lang_identifier = LangIdentifier()
identify_lang = lang_identifier.identify_lang
